# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import click

from threading import Lock
from typing import cast
from typing_extensions import override
from collections.abc import Iterable
from uuid import UUID
from sqlalchemy import text

from odc.geo import Geometry, CRS
from datacube import Datacube
from datacube.model import Product, Dataset

from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.index.api import OWSAbstractIndex, OWSAbstractIndexDriver, LayerSignature, LayerExtent, TimeSearchTerm
from .product_ranges import create_range_entry as create_range_entry_impl, get_ranges as get_ranges_impl
from .mv_index import MVSelectOpts, mv_search
from datacube_ows.index.sql import run_sql


class OWSPostgresIndex(OWSAbstractIndex):
    name: str = "postgres"

    # method to check database access (for ping op)
    @override
    def check_db_access(self, dc: Datacube) -> bool:
        db_ok = False
        try:
            with dc.index._db.give_me_a_connection() as conn:  # type: ignore[attr-defined]
                results = conn.execute(text("""
                    SELECT *
                    FROM ows.layer_ranges
                    LIMIT 1""")
                )
                for r in results:
                    db_ok = True
        except:
            pass
        return db_ok

    # method to delete obsolete schemas etc.
    @override
    def cleanup_schema(self, dc: Datacube) -> None:
        self._run_sql(dc, "ows_schema/cleanup")

    # Schema creation method
    @override
    def create_schema(self, dc: Datacube) -> None:
        click.echo("Creating/updating schema and tables...")
        self._run_sql(dc, "ows_schema/create")
        click.echo("Creating/updating materialised views...")
        self._run_sql(dc, "extent_views/create")
        click.echo("Setting ownership of materialised views...")
        self._run_sql(dc, "extent_views/grants/refresh_owner")

    # Permission management method
    @override
    def grant_perms(self, dc: Datacube, role: str, read_only: bool = False) -> None:
        if read_only:
            self._run_sql(dc, "ows_schema/grants/read_only", role=role)
            self._run_sql(dc, "extent_views/grants/read_only", role=role)
        else:
            self._run_sql(dc, "ows_schema/grants/read_write", role=role)
            self._run_sql(dc, "extent_views/grants/write_refresh", role=role)

    # Spatiotemporal index update method (e.g. refresh materialised views)
    @override
    def update_geotemporal_index(self, dc: Datacube) -> None:
        self._run_sql(dc, "extent_views/refresh")

    @override
    def create_range_entry(self, layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]) -> None:
        create_range_entry_impl(layer, cache)

    @override
    def get_ranges(self, layer: OWSNamedLayer) -> LayerExtent | None:
        return get_ranges_impl(layer)

    @override
    def ds_search(self,
                  layer: OWSNamedLayer,
                  times: Iterable[TimeSearchTerm] | None = None,
                  geom: Geometry | None = None,
                  products: Iterable[Product] | None = None
                  ) -> Iterable[Dataset]:
        return cast(Iterable[Dataset], mv_search(layer.dc.index, MVSelectOpts.DATASETS,
                                                 times=times, geom=geom, products=products))

    @override
    def dsid_search(self,
                    layer: OWSNamedLayer,
                    times: Iterable[TimeSearchTerm] | None = None,
                    geom: Geometry | None = None,
                    products: Iterable[Product] | None = None
                    ) -> Iterable[UUID]:
        return cast(Iterable[UUID], mv_search(layer.dc.index, MVSelectOpts.IDS,
                                              times=times, geom=geom, products=products))

    @override
    def count(self,
              layer: OWSNamedLayer,
              times: Iterable[TimeSearchTerm] | None = None,
              geom: Geometry | None = None,
              products: Iterable[Product] | None = None
              ) -> int:
        return cast(int, mv_search(layer.dc.index, MVSelectOpts.COUNT,
                                   times=times, geom=geom, products=products))

    @override
    def extent(self,
               layer: OWSNamedLayer,
               times: Iterable[TimeSearchTerm] | None = None,
               geom: Geometry | None = None,
               products: Iterable[Product] | None = None,
               crs: CRS | None = None
               ) -> Geometry | None:
        extent = cast(Geometry | None, mv_search(layer.dc.index, MVSelectOpts.EXTENT,
                                                 times=times, geom=geom, products=products))
        if extent is None or crs is None or crs == extent.crs:
            return extent
        else:
            return extent.to_crs(crs)

    def _run_sql(self, dc: Datacube, path: str, **params: str) -> bool:
        return run_sql(dc, self.name, path, **params)


pgdriverlock = Lock()


class OWSPostgresIndexDriver(OWSAbstractIndexDriver):
    _driver = None
    @classmethod
    @override
    def ows_index_class(cls) -> type[OWSAbstractIndex]:
        return OWSPostgresIndex

    @classmethod
    @override
    def ows_index(cls) -> OWSAbstractIndex:
        with pgdriverlock:
            if cls._driver is None:
                cls._driver = OWSPostgresIndex()
        return cls._driver


def ows_index_driver_init() -> OWSPostgresIndexDriver:
    return OWSPostgresIndexDriver()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

from datacube_ows.styles.base import StyleDefBase  # noqa: F401
from datacube_ows.styles.colormap import ColorMapStyleDef  # noqa: F401
from datacube_ows.styles.component import ComponentStyleDef  # noqa: F401
from datacube_ows.styles.hybrid import HybridStyleDef  # noqa: F401
from datacube_ows.styles.ramp import ColorRampDef  # noqa: F401

StyleDef = StyleDefBase



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import math
from datetime import date, datetime, timezone
from collections.abc import Callable
import click

import datacube
import odc.geo
import sqlalchemy.exc
from psycopg2.extras import Json
from sqlalchemy import text

from odc.geo.geom import CRS

from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.utils import get_sqlconn
from datacube_ows.index.api import CoordRange, LayerSignature, LayerExtent

_LOG = logging.getLogger(__name__)


def jsonise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    return {
        "top": bbox.top,
        "bottom": bbox.bottom,
        "left": bbox.left,
        "right": bbox.right,
    }


def create_range_entry(layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]) -> None:
    meta = LayerSignature(time_res=layer.time_resolution.value,
                          products=tuple(layer.product_names),
                          env=layer.local_env._name,
                          datasets=layer.dc.index.datasets.count(product=layer.product_names))

    click.echo(f"Postgis Updating range for layer {layer.name}")
    click.echo(f"(signature: {meta.as_json()!r})")
    conn = get_sqlconn(layer.dc)
    txn = conn.begin()
    if meta in cache:
        template = cache[meta][0]
        click.echo(f"Layer {template} has same signature - reusing")
        cache[meta].append(layer.name)
        try:
            conn.execute(text("""
            INSERT INTO ows.layer_ranges
                (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
            SELECT :layer_id, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated
            FROM ows.layer_ranges lr2
            WHERE lr2.layer = :template_id"""),
                         {
                                      "layer_id": layer.name,
                                      "template_id": template
                                   })
        except sqlalchemy.exc.IntegrityError:
            conn.execute(text("""
            UPDATE ows.layer_ranges lr1
            SET lat_min = lr2.lat_min,
                lat_max = lr2.lat_max,
                lon_min = lr2.lon_min,
                lon_max = lr2.lon_max,
                dates = lr2.dates,
                bboxes = lr2.bboxes,
                meta = lr2.meta,
                last_updated = lr2.last_updated
            FROM ows.layer_ranges lr2
            WHERE lr1.layer = :layer_id
            AND   lr2.layer = :template_id"""),
                         {
                             "layer_id": layer.name,
                             "template_id": template
                         })
    else:
        # insert empty row if one does not already exist
        conn.execute(text("""
        INSERT INTO ows.layer_ranges
        (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
        VALUES
        (:p_layer, 0, 0, 0, 0, :empty, :empty, :meta, :now)
        ON CONFLICT (layer) DO NOTHING
        """),
        {
            "p_layer": layer.name, "empty": Json(""),
            "meta": Json(meta.as_json()), "now": datetime.now(tz=timezone.utc)
        })

        prodids = [p.id for p in layer.products]

        # Set default timezone
        conn.execute(text("""set timezone to 'Etc/UTC'"""))

        # Loop over dates
        dates = set()  # Should get to here!
        if layer.time_resolution.is_solar():
          results = conn.execute(text(
              """
              select lower(dt.search_val), upper(dt.search_val),
                     lower(lon.search_val), upper(lon.search_val)
              from odc.dataset ds, odc.dataset_search_datetime dt, odc.dataset_search_num lon
              where ds.product_ref = ANY(:prodids)
              AND ds.id = dt.dataset_ref
              AND ds.id = lon.dataset_ref
              AND dt.search_key = :time
              AND lon.search_key = :lon
              """),
              {"prodids": prodids, "time": "time", "lon": "lon"})
          for result in results:
              dt1, dt2, ll, lu = result
              lon = (ll + lu) / 2
              dt = dt1 + (dt2 - dt1) / 2
              dt = dt.astimezone(timezone.utc)

              solar_day = datacube.api.query._convert_to_solar_time(dt, lon).date()
              dates.add(solar_day)
        else:
          results = conn.execute(text(
              """
              select
                    array_agg(dt.search_val)
              from odc.dataset_search_datetime dt,
                   odc.dataset ds
              WHERE ds.product_ref = ANY(:prodids)
              AND   ds.id = dt.dataset_ref
              AND   dt.search_key = 'time'
              """),
              {"prodids": prodids}
          )
          for result in results:
              for dat_ran in result[0]:
                  dates.add(dat_ran.lower)

        if layer.time_resolution.is_subday():
          date_formatter = lambda d: d.isoformat()
        else:
          date_formatter = lambda d: d.strftime("%Y-%m-%d")

        dates = sorted(dates)
        conn.execute(text("""
           UPDATE ows.layer_ranges
           SET dates = :dates
           WHERE layer= :layer_id
        """),
                   {
                       "dates": Json(list(map(date_formatter, dates))),
                       "layer_id": layer.name
                   }
        )
        # calculate bounding boxes
        # Get extent polygon from materialised views

        all_bboxes = build_bboxes(layer)
        if not all_bboxes:
            return

        conn.execute(text("""
        UPDATE ows.layer_ranges
        SET bboxes = :bbox,
            lat_min = :lat_min,
            lat_max = :lat_max,
            lon_min = :lon_min,
            lon_max = :lon_max
        WHERE layer = :layer_id
        """), {
            "bbox": Json(all_bboxes),
            "layer_id": layer.name,
            "lat_min": all_bboxes['EPSG:4326']['bottom'],
            "lat_max": all_bboxes['EPSG:4326']['top'],
            "lon_min": all_bboxes['EPSG:4326']['left'],
            "lon_max": all_bboxes['EPSG:4326']['right']
        })

        cache[meta] = [layer.name]

    txn.commit()
    conn.close()


def build_bboxes(layer: OWSNamedLayer) -> dict[str, dict[str, float]]:
    # User native CRS for base extent, unless no spatial index for it, in which case
    # use 4326 which always has a spatial index.
    base_crs = CRS(layer.native_CRS)
    if base_crs not in layer.dc.index.spatial_indexes():
        base_crs = CRS("epsg:4326")
    base_extent = extent_for_layer(layer, base_crs)
    if base_extent is None:
        click.echo(f"Layer {layer.name} has no extent in CRS {base_crs}.  Skipping.")
        return {}
    result = {}
    for crsid, crs in layer.global_config().crses.items():
        if crs == base_crs:
            result[crsid] = sanitise_bbox(base_extent.boundingbox)
        elif crs in layer.dc.index.spatial_indexes():
            extent = extent_for_layer(layer, crs)
            if extent is None:
                click.echo(f"Layer {layer.name} has no extent in CRS {crs}.  Skipping.")
                return {}
            result[crsid] = sanitise_bbox(extent.boundingbox)
        else:
            extent = base_extent.to_crs(crs)
            result[crsid] = sanitise_bbox(extent.boundingbox)
    return result


def extent_for_layer(layer: OWSNamedLayer, crs: CRS) -> odc.geo.Geometry | None:
    ext = None
    for product in layer.products:
        prod_extent = layer.dc.index.products.spatial_extent(product, crs)
        if ext is None:
            ext = prod_extent
        else:
            ext = ext | prod_extent
    return ext


def bbox_projections(starting_box: odc.geo.Geometry, crses: dict[str, odc.geo.CRS]) -> dict[str, dict[str, float]]:
   result = {}
   for crsid, crs in crses.items():
        projbbox = starting_box.to_crs(crs).boundingbox
        result[crsid] = sanitise_bbox(projbbox)
   return result


def sanitise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    def sanitise_coordinate(coord: float, fallback: float, upper: bool) -> float:
        if not math.isfinite(coord):
            return fallback
        elif upper and coord > fallback:
            return fallback
        elif not upper and coord < fallback:
            return fallback
        else:
            return coord
    if bbox.crs == CRS("epsg:4326"):
        return {
            "top": sanitise_coordinate(bbox.top, float("90.0"), True),
            "bottom": sanitise_coordinate(bbox.bottom, float("-90.0"), False),
            "left": sanitise_coordinate(bbox.left, float("-180"), False),
            "right": sanitise_coordinate(bbox.right, float("9.999999999e99"), True),
        }
    else:
        return {
            "top": sanitise_coordinate(bbox.top, float("9.999999999e99"), True),
            "bottom": sanitise_coordinate(bbox.bottom, float("-9.999999999e99"), False),
            "left": sanitise_coordinate(bbox.left, float("-9.999999999e99"), False),
            "right": sanitise_coordinate(bbox.right, float("9.999999999e99"), True),
        }


def get_ranges(layer: OWSNamedLayer) -> LayerExtent | None:
    cfg = layer.global_cfg
    conn = get_sqlconn(layer.dc)
    results = conn.execute(text("""
        SELECT *
        FROM ows.layer_ranges
        WHERE layer=:pname"""),
                           {"pname": layer.name}
                          )

    for result in results:
        conn.close()
        if layer.time_resolution.is_subday():
            dt_parser: Callable[[str], datetime | date] = lambda dts: datetime.fromisoformat(dts)
        else:
            dt_parser = lambda dts: datetime.strptime(dts, "%Y-%m-%d").date()
        times = [dt_parser(d) for d in result.dates if d is not None]
        if not times:
            return None
        return LayerExtent(
            lat=CoordRange(min=float(result.lat_min), max=float(result.lat_max)),
            lon=CoordRange(min=float(result.lon_min), max=float(result.lon_max)),
            times=times,
            bboxes=result.bboxes
        )
    return None



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

from typing import cast
from typing_extensions import override

from xarray import DataArray, Dataset

from datacube_ows.config_utils import CFG_DICT, ConfigException
from datacube_ows.styles.base import StyleDefBase
from datacube_ows.styles.component import LINEAR_COMP_DICT, ComponentStyleDef
from datacube_ows.styles.ramp import ColorRampDef

TYPE_CHECKING = False
if TYPE_CHECKING:
    from datacube_ows.ows_configuration import OWSNamedLayer


class HybridStyleDef(ColorRampDef, ComponentStyleDef):
    """
    Hybrid component/colour ramp style type.

    Returns a linear blend of a component image and colour ramp image
    """
    auto_legend = False

    def __init__(self,
                 product: "OWSNamedLayer",
                 style_cfg: CFG_DICT,
                 defer_multi_date: bool = False,
                 stand_alone: bool = False,
                 user_defined: bool = False) -> None:
        """
        See StyleBaseDef
        """
        super().__init__(product, style_cfg,
                                 defer_multi_date=defer_multi_date,
                                 stand_alone=stand_alone,
                                 user_defined=user_defined)
        style_cfg = cast(CFG_DICT, self._raw_cfg)
        self.component_ratio = float(cast(float | str, style_cfg["component_ratio"]))
        if self.component_ratio < 0.0 or self.component_ratio > 1.0:
            raise ConfigException("Component ratio must be a floating point number between 0 and 1")

    @override
    def transform_single_date_data(self, data: Dataset) -> Dataset:
        """
        Apply style to raw data to make an RGBA image xarray (single time slice only)

        :param data: Raw data, all bands.
        :return: RGBA uint8 xarray
        """
        #pylint: disable=too-many-locals
        if self.index_function is not None:
            data['index_function'] = (data.dims, self.index_function(data).data)

        imgdata = Dataset(coords=data)

        d: DataArray = data['index_function']
        for band, intensity in self.rgb_components.items():
            rampdata = DataArray(self.color_ramp.get_value(d, band),
                                 coords=d.coords,
                                 dims=d.dims)
            component_band_data: DataArray | None = None
            for c_band, c_intensity in cast(LINEAR_COMP_DICT, self.rgb_components[band]).items():
                if callable(c_intensity):
                    imgband_component_data = cast(DataArray, c_intensity(data[c_band], c_band, band))
                else:
                    imgband_component_data = data[c_band] * cast(DataArray, c_intensity)
                if component_band_data is not None:
                    component_band_data += imgband_component_data
                else:
                    component_band_data = imgband_component_data
                if band != "alpha":
                    component_band_data = self.compress_band(band, component_band_data)
            img_band_data = (rampdata * 255.0 * (1.0 - self.component_ratio)
                             + self.component_ratio * cast(DataArray,
                                                           component_band_data))
            imgdata[band] = (d.dims, img_band_data.astype("uint8").data)

        return imgdata

# Register HybridStyleDef as a priority subclass of StyleDefBase
#    (priority means takes precedence over ComponentStyleDef and ColourRampDef)
StyleDefBase.register_subclass(HybridStyleDef, "component_ratio", priority=True)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0


style_s2_pure_aerosol = {
    "name": "aerosol",
    "title": "Narrow Blue - 440",
    "abstract": "Coastal Aerosol or Narrow Blue band, approximately 435nm to 450nm",
    "components": {
        "red": {"coastal_aerosol": 1.0},
        "green": {"coastal_aerosol": 1.0},
        "blue": {"coastal_aerosol": 1.0},
    },
    "scale_range": [0.0, 3000.0],
}


style_s2_pure_blue = {
    "name": "blue",
    "title": "Blue - 490",
    "abstract": "Blue band, approximately 453nm to 511nm",
    "components": {"red": {"blue": 1.0}, "green": {"blue": 1.0}, "blue": {"blue": 1.0}},
    "scale_range": [0.0, 3000.0],
}

style_s2_pure_green = {
    "name": "green",
    "title": "Green - 560",
    "abstract": "Green band, approximately 534nm to 588nm",
    "components": {
        "red": {"green": 1.0},
        "green": {"green": 1.0},
        "blue": {"green": 1.0},
    },
    "scale_range": [0.0, 3000.0],
}


style_s2_pure_red = {
    "name": "red",
    "title": "Red - 670",
    "abstract": "Red band, roughly 637nm to 672nm",
    "components": {"red": {"red": 1.0}, "green": {"red": 1.0}, "blue": {"red": 1.0}},
    "scale_range": [0.0, 3000.0],
}


style_s2_pure_redge_1 = {
    "name": "red_edge_1",
    "title": "Vegetation Red Edge - 710",
    "abstract": "Near infra-red band, centred on 710nm",
    "components": {
        "red": {"red_edge_1": 1.0},
        "green": {"red_edge_1": 1.0},
        "blue": {"red_edge_1": 1.0},
    },
    "scale_range": [0.0, 3000.0],
}


style_s2_pure_redge_2 = {
    "name": "red_edge_2",
    "title": "Vegetation Red Edge - 740",
    "abstract": "Near infra-red band, centred on 740nm",
    "components": {
        "red": {"red_edge_2": 1.0},
        "green": {"red_edge_2": 1.0},
        "blue": {"red_edge_2": 1.0},
    },
    "scale_range": [0.0, 3000.0],
}


style_s2_pure_redge_3 = {
    "name": "red_edge_3",
    "title": "Vegetation Red Edge - 780",
    "abstract": "Near infra-red band, centred on 780nm",
    "components": {
        "red": {"red_edge_3": 1.0},
        "green": {"red_edge_3": 1.0},
        "blue": {"red_edge_3": 1.0},
    },
    "scale_range": [0.0, 3000.0],
}


style_s2_pure_nir = {
    "name": "nir",
    "title": "Near Infrared (NIR) - 840",
    "abstract": "Near infra-red band, roughly 853nm to 876nm",
    "components": {"red": {"nir": 1.0}, "green": {"nir": 1.0}, "blue": {"nir": 1.0}},
    "scale_range": [0.0, 3000.0],
}

style_s2_pure_narrow_nir = {
    "name": "narrow_nir",
    "title": "Narrow Near Infrared - 870",
    "abstract": "Near infra-red band, centred on 865nm",
    "components": {"red": {"nir": 1.0}, "green": {"nir": 1.0}, "blue": {"nir": 1.0}},
    "scale_range": [0.0, 3000.0],
}


style_s2_pure_swir1 = {
    "name": "swir1",
    "title": "Shortwave Infrared (SWIR) - 1610",
    "abstract": "Short wave infra-red band 1, roughly 1575nm to 1647nm",
    "components": {
        "red": {"swir_2": 1.0},
        "green": {"swir_2": 1.0},
        "blue": {"swir_2": 1.0},
    },
    "scale_range": [0.0, 3000.0],
}


style_s2_pure_swir2 = {
    "name": "swir2",
    "title": "Shortwave Infrared (SWIR) - 2190",
    "abstract": "Short wave infra-red band 2, roughly 2117nm to 2285nm",
    "components": {
        "red": {"swir_3": 1.0},
        "green": {"swir_3": 1.0},
        "blue": {"swir_3": 1.0},
    },
    "scale_range": [0.0, 3000.0],
}



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import numpy

# Style index functions


def scale_data(imgband_data, scale_from, scale_to):
    sc_min, sc_max = scale_from
    tc_min, tc_max = scale_to
    clipped = imgband_data.clip(sc_min, sc_max)
    normalised = (clipped - sc_min) / (sc_max - sc_min)
    scaled = normalised * (tc_max - tc_min)
    return scaled + tc_min


def scalable(undecorated):
    def decorated(*args, **kwargs):
        scale_from = kwargs.pop("scale_from", None)
        scale_to = kwargs.pop("scale_to", None)
        if scale_from is not None and scale_to is None:
            scale_to = [0, 255]
        unscaled = undecorated(*args, **kwargs)
        if scale_from:
            return scale_data(unscaled, scale_from, scale_to)
        return unscaled

    return decorated


def band_modulator(undecorated):
    def decorated(data, *args, **kwargs):
        band_mapper = kwargs.get("band_mapper", None)
        mult_band = kwargs.pop("mult_band", None)
        raw_data = undecorated(data, *args, **kwargs)
        if mult_band and band_mapper:
            mult_band = band_mapper(mult_band)
        if mult_band:
            return data[mult_band] * raw_data
        return raw_data
    return decorated

def pre_scaled_band(data, band, scale, offset):
    # Pre-scale a band as `data[band] * scale + offset`
    return data[band] * scale + offset

def sum_bands(data, band1, band2, band_mapper=None):
    if band_mapper:
        band1 = band_mapper(band1)
        band2 = band_mapper(band2)
    return data[band1] + data[band2]

def pre_scaled_sum_bands(
    data, band1, band2,
    scale1=1.0, offset1=0.0,
    scale2=1.0, offset2=0.0,
    band_mapper=None):
    # Calculate the sum of two bands, after pre-scaling them with a scale and offset
    if band_mapper:
        band1 = band_mapper(band1)
        band2 = band_mapper(band2)
    # Pre-scaled bands data
    data1 = pre_scaled_band(data, band1, scale1, offset1)
    data2 = pre_scaled_band(data, band2, scale2, offset2)
    return data1 + data2


def delta_bands(data, band1, band2, band_mapper=None):
    if band_mapper:
        band1 = band_mapper(band1)
        band2 = band_mapper(band2)
    typ1 = data[band1].dtype
    typ2 = data[band2].dtype
    if typ1.name.startswith('uint'):
        nodata = data[band1].nodata
        data[band1] = data[band1].astype('int32')
        data[band1].attrs["nodata"] = nodata
    if typ2.name.startswith('uint'):
        nodata = data[band2].nodata
        data[band2] = data[band2].astype('int32')
        data[band2].attrs["nodata"] = nodata
    # if typ1.name.startswith('uint') or typ2.name.startswith('uint'):
        # data = data.astype('int32', copy=False)
    return data[band1] - data[band2]

def pre_scaled_delta_bands(
    data, band1, band2,
    scale1=1.0, offset1=0.0,
    scale2=1.0, offset2=0.0,
    band_mapper=None):
    # Calculate the difference between two bands, after pre-scaling them with a scale
    # and offset
    if band_mapper:
        band1 = band_mapper(band1)
        band2 = band_mapper(band2)
    typ1 = data[band1].dtype
    typ2 = data[band2].dtype
    if typ1.name.startswith('uint'):
        nodata = data[band1].nodata
        data[band1] = data[band1].astype('int32')
        data[band1].attrs["nodata"] = nodata
    if typ2.name.startswith('uint'):
        nodata = data[band2].nodata
        data[band2] = data[band2].astype('int32')
        data[band2].attrs["nodata"] = nodata
    # if typ1.name.startswith('uint') or typ2.name.startswith('uint'):
        # data = data.astype('int32', copy=False)
        # Pre-scaled bands data
    data1 = pre_scaled_band(data, band1, scale1, offset1)
    data2 = pre_scaled_band(data, band2, scale2, offset2)
    return data1 - data2

# N.B. Modifying scale_to would be dangerous - don't do it.
# pylint: disable=dangerous-default-value
@scalable
def norm_diff(data, band1, band2, band_mapper=None):
    # Calculate a normalised difference index.
    return delta_bands(data, band1, band2, band_mapper) / sum_bands(data, band1, band2, band_mapper)

@scalable
def pre_scaled_norm_diff(data, band1, band2,
                         scale1=1.0, offset1=0.0,
                         scale2=1.0, offset2=0.0,
                         band_mapper=None,):
    # Calculate a normalised difference index, after scaling the input bands with a
    # scale and offset
    return (pre_scaled_delta_bands(
        data, band1, band2, scale1, offset1, scale2, offset2, band_mapper) /
        pre_scaled_sum_bands(
            data, band1, band2, scale1, offset1, scale2, offset2, band_mapper)
    )
@scalable
def constant(data, band, const, band_mapper=None):
    # Covert an xarray for a flat constant.
    # Useful for displaying mask extents as a flat colour and debugging.
    # params is assumed to be a tuple containing a constant value and a band name/alias.

    if band_mapper:
        band = band_mapper(band)
    return data[band] * 0.0 + const


@scalable
def single_band(data, band, band_mapper=None):
    # Use the raw value of a band directly as the index function.

    if band_mapper:
        band = band_mapper(band)
    return data[band]


@scalable
def band_quotient(data, band1, band2, band_mapper=None):
    if band_mapper:
        band1 = band_mapper(band1)
        band2 = band_mapper(band2)
    return data[band1] / data[band2]


@scalable
def band_quotient_sum(data, band1a, band1b, band2a, band2b, band_mapper=None):
    return band_quotient(data, band1a, band1b, band_mapper) + band_quotient(data, band2a, band2b, band_mapper)


@scalable
def sentinel2_ndci(data, b_red_edge, b_red, b_green, b_swir, band_mapper=None):
    red_delta = delta_bands(data, b_red_edge, b_red, band_mapper)
    red_sum = sum_bands(data, b_red_edge, b_red, band_mapper)
    mndwi = norm_diff(data, b_green, b_swir, band_mapper)

    return red_delta / red_sum.where(mndwi > 0.1)


def multi_date_delta(data, time_direction=-1):
    data1, data2 = (data.sel(time=dt) for dt in data.coords["time"].values)

#    data1, data2 = data.values.item(0), data.values.item(1)
    if time_direction >= 0:
        return data1 - data2
    else:
        return data2 - data1

def multi_date_pass(data):
    return data

def multi_date_raw_example(data, band1, band2, band_mapper=None):
    if band_mapper:
        band1 = band_mapper(band1)
        band2 = band_mapper(band2)
    data1, data2 = (data.sel(time=dt) for dt in data.coords["time"].values)
    return (
            (data2[band1] - data1[band1]) - (data2[band2] - data1[band2])
    ) / (
            data1[band1] + data1[band2] + data2[band1] + data2[band2]
    )


@band_modulator
@scalable
def single_band_log(data, band, scale_factor, exponent, band_mapper=None):
    if band_mapper:
        band = band_mapper(band)
    d = data[band]
    return scale_factor * ((d ** exponent) - 1.0)


@band_modulator
@scalable
def single_band_arcsec(data, band, band_mapper=None):
    if band_mapper:
        band = band_mapper(band)
    d = data[band]
    return numpy.arccos(1 / (d + 1))


@band_modulator
@scalable
def single_band_offset_log(data, band, scale=1.0, offset=None, band_mapper=None):
    if band_mapper:
        band = band_mapper(band)
    d = data[band]
    if offset is not None:
        d = data[band] * scale + offset
        unscaled = numpy.log(d)
    else:
        unscaled = numpy.log1p(d * scale)
    return unscaled


@scalable
def radar_vegetation_index(data, band_hv, band_hh, band_mapper=None):
    if band_mapper:
        band_hv = band_mapper(band_hv)
        band_hh = band_mapper(band_hh)
    hv_sq = data[band_hv] * data[band_hv]
    hh_sq = data[band_hh] * data[band_hh]
    return (hv_sq * 4.0) / (hh_sq + hv_sq)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0


from local_cfg.africa_ls_sr_local_cfg.bands import (reslim_landsat,
                                                            bands_ls5_sr,
                                                            bands_ls7_sr,
                                                            bands_ls8_sr)
from local_cfg.africa_ls_sr_local_cfg.styles import (
                                                styles_ls8c2_sr_list, styles_lsc2_sr_list)

layer_ls8 = {
    "title": "Surface Reflectance Landsat 8 (USGS Collection 2)",
    "name": "ls8_sr",
    "abstract": """
Surface reflectance is the fraction of incoming solar radiation that is reflected from Earth's surface. Variations in satellite measured radiance due to atmospheric properties have been corrected for, so images acquired over the same area at different times are comparable and can be used readily to detect changes on Earth’s surface.

DE Africa provides access to Landsat Collection 2 Level-2 Surface Reflectance products over Africa. USGS Landsat Collection 2 offers improved processing, geometric accuracy, and radiometric calibration compared to previous Collection 1 products. The Level-2 products are endorsed by the Committee on Earth Observation Satellites (CEOS) to be Analysis Ready Data for Land (CARD4L)-compliant.

More techincal information about the Landsat Surface Reflectance product can be found in the User Guide (https://docs.digitalearthafrica.org/en/latest/data_specs/Landsat_C2_SR_specs.html).

Landsat 8 product has a spatial resolution of 30 m and a temporal coverage of 2013 to present.

Landsat Level- 2 Surface Reflectance Science Product courtesy of the U.S. Geological Survey.

For more information on Landsat products, see https://www.usgs.gov/core-science-systems/nli/landsat/landsat-collection-2-level-2-science-products.

This product is accessible through OGC Web Service (https://ows.digitalearth.africa/), for analysis in DE Africa Sandbox JupyterLab (https://github.com/digitalearthafrica/deafrica-sandbox-notebooks/wiki) and for direct download from AWS S3 (https://data.digitalearth.africa/).
""",
    "product_name": "ls8_sr",
    "bands": bands_ls8_sr,
    "dynamic": True,
    "resource_limits": reslim_landsat,
    "image_processing": {
        "extent_mask_func": "datacube_ows.ogc_utils.mask_by_val",
        "always_fetch_bands": [],
        "manual_merge": False,  # True
        "apply_solar_corrections": False,
    },
    "native_crs": "EPSG:4326",
    "native_resolution": [30.0, -30.0],
    "wcs": {
        "default_bands": ["red", "green", "blue"],
    },
    "styling": {
        "default_style": "simple_rgb",
        "styles": styles_ls8c2_sr_list,
    },
}

layer_ls7 = {
    "title": "Surface Reflectance Landsat 7 (USGS Collection 2)",
    "name": "ls7_sr",
    "abstract": """
Surface reflectance is the fraction of incoming solar radiation that is reflected from Earth's surface. Variations in satellite measured radiance due to atmospheric properties have been corrected for, so images acquired over the same area at different times are comparable and can be used readily to detect changes on Earth’s surface.

DE Africa provides access to Landsat Collection 2 Level-2 Surface Reflectance products over Africa. USGS Landsat Collection 2 offers improved processing, geometric accuracy, and radiometric calibration compared to previous Collection 1 products. The Level-2 products are endorsed by the Committee on Earth Observation Satellites (CEOS) to be Analysis Ready Data for Land (CARD4L)-compliant.

More techincal information about the Landsat Surface Reflectance product can be found in the User Guide (https://docs.digitalearthafrica.org/en/latest/data_specs/Landsat_C2_SR_specs.html).

Landsat 7 product has a spatial resolution of 30 m and a temporal coverage of 1999 to present.

Landsat Level- 2 Surface Reflectance Science Product courtesy of the U.S. Geological Survey.

For more information on Landsat products, see https://www.usgs.gov/core-science-systems/nli/landsat/landsat-collection-2-level-2-science-products.

This product is accessible through OGC Web Service (https://ows.digitalearth.africa/), for analysis in DE Africa Sandbox JupyterLab (https://github.com/digitalearthafrica/deafrica-sandbox-notebooks/wiki) and for direct download from AWS S3 (https://data.digitalearth.africa/).
""",
    "product_name": "ls7_sr",
    "bands": bands_ls7_sr,
    "dynamic": True,
    "resource_limits": reslim_landsat,
    "image_processing": {
        "extent_mask_func": "datacube_ows.ogc_utils.mask_by_val",
        "always_fetch_bands": [],
        "manual_merge": False,  # True
        "apply_solar_corrections": False,
    },
    "native_crs": "EPSG:4326",
    "native_resolution": [30.0, -30.0],
    "wcs": {
        "default_bands": ["red", "green", "blue"],
    },
    "styling": {
        "default_style": "simple_rgb",
        "styles": styles_lsc2_sr_list,
    },
}

layer_ls5 = {
    "title": "Surface Reflectance Landsat 5 (USGS Collection 2)",
    "name": "ls5_sr",
    "abstract": """
Surface reflectance is the fraction of incoming solar radiation that is reflected from Earth's surface. Variations in satellite measured radiance due to atmospheric properties have been corrected for, so images acquired over the same area at different times are comparable and can be used readily to detect changes on Earth’s surface.

DE Africa provides access to Landsat Collection 2 Level-2 Surface Reflectance products over Africa. USGS Landsat Collection 2 offers improved processing, geometric accuracy, and radiometric calibration compared to previous Collection 1 products. The Level-2 products are endorsed by the Committee on Earth Observation Satellites (CEOS) to be Analysis Ready Data for Land (CARD4L)-compliant.

More techincal information about the Landsat Surface Reflectance product can be found in the User Guide (https://docs.digitalearthafrica.org/en/latest/data_specs/Landsat_C2_SR_specs.html).

Landsat 5 product has a spatial resolution of 30 m and a temporal coverage of 1984 to 2012.

Landsat Level- 2 Surface Reflectance Science Product courtesy of the U.S. Geological Survey.

For more information on Landsat products, see https://www.usgs.gov/core-science-systems/nli/landsat/landsat-collection-2-level-2-science-products.

This product is accessible through OGC Web Service (https://ows.digitalearth.africa/), for analysis in DE Africa Sandbox JupyterLab (https://github.com/digitalearthafrica/deafrica-sandbox-notebooks/wiki) and for direct download from AWS S3 (https://data.digitalearth.africa/).
""",
    "product_name": "ls5_sr",
    "bands": bands_ls5_sr,
    "dynamic": True,
    "resource_limits": reslim_landsat,
    "image_processing": {
        "extent_mask_func": "datacube_ows.ogc_utils.mask_by_val",
        "always_fetch_bands": [],
        "manual_merge": False,  # True
        "apply_solar_corrections": False,
    },
    "native_crs": "EPSG:4326",
    "native_resolution": [30.0, -30.0],
    "wcs": {
        "default_bands": ["red", "green", "blue"],
    },
    "styling": {
        "default_style": "simple_rgb",
        "styles": styles_lsc2_sr_list,
    },
}



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import dataclasses
from abc import ABC, abstractmethod
from datetime import datetime, date
from typing import NamedTuple
from collections.abc import Iterable
from uuid import UUID

from datacube import Datacube
from datacube.index.abstract import AbstractIndex
from datacube.model import Product, Dataset
from odc.geo.geom import Geometry, CRS, polygon

from datacube_ows.config_utils import CFG_DICT, ConfigException

TYPE_CHECKING = False
if TYPE_CHECKING:
    from datacube_ows.ows_configuration import OWSNamedLayer

class AbortRun(Exception):
    pass


@dataclasses.dataclass(frozen=True)
class LayerSignature:
    time_res: str
    products: tuple[str, ...]
    env: str
    datasets: int

    def as_json(self) -> dict[str, list[str] | str | int]:
        return {
            "time_res": self.time_res,
            "products": list(self.products),
            "env": self.env,
            "datasets": self.datasets,
        }


DateOrDateTime = datetime | date
TimeSearchTerm = tuple[datetime, datetime] | tuple[date, date] | DateOrDateTime


class CoordRange(NamedTuple):
    min: float
    max: float


class LayerExtent:
    def __init__(self, lat: CoordRange, lon: CoordRange, times: list[DateOrDateTime], bboxes: CFG_DICT) -> None:
        self.lat = lat
        self.lon = lon
        self.times = times
        self.start_time = times[0]
        self.end_time = times[-1]
        self.time_set = set(times)
        self.bboxes = bboxes


class OWSAbstractIndex(ABC):
    name: str = ""

    # method to check database access (for ping op)
    @abstractmethod
    def check_db_access(self, dc: Datacube) -> bool:
        ...

    # method to delete obsolete schemas etc.
    @abstractmethod
    def cleanup_schema(self, dc: Datacube):
        ...

    # Schema creation method
    @abstractmethod
    def create_schema(self, dc: Datacube):
        ...

    # Permission management method
    @abstractmethod
    def grant_perms(self, dc: Datacube, role: str, read_only: bool = False):
        ...

    # Spatiotemporal index update method (e.g. refresh materialised views)
    @abstractmethod
    def update_geotemporal_index(self, dc: Datacube):
        ...

    # Range table update method
    @abstractmethod
    def create_range_entry(self, layer: "OWSNamedLayer", cache: dict[LayerSignature, list[str]]) -> None:
        ...

    # Range table read method
    @abstractmethod
    def get_ranges(self, layer: "OWSNamedLayer") -> LayerExtent | None:
        ...

    # Spatiotemporal search methods
    @abstractmethod
    def ds_search(self,
                  layer: "OWSNamedLayer",
                  times: Iterable[TimeSearchTerm] | None = None,
                  geom: Geometry | None = None,
                  products: Iterable[Product] | None = None
                  ) -> Iterable[Dataset]:
        ...

    def dsid_search(self,
                    layer: "OWSNamedLayer",
                    times: Iterable[TimeSearchTerm] | None = None,
                    geom: Geometry | None = None,
                    products: Iterable[Product] | None = None
                    ) -> Iterable[UUID]:
        for ds in self.ds_search(layer, times, geom, products):
            yield ds.id

    def count(self,
              layer: "OWSNamedLayer",
              times: Iterable[TimeSearchTerm] | None = None,
              geom: Geometry | None = None,
              products: Iterable[Product] | None = None
              ) -> int:
        return len([dsid for dsid in self.dsid_search(layer, times, geom, products)])

    def extent(self,
               layer: "OWSNamedLayer",
               times: Iterable[TimeSearchTerm] | None = None,
               geom: Geometry | None = None,
               products: Iterable[Product] | None = None,
               crs: CRS | None = None
               ) -> Geometry | None:
        geom = self._prep_geom(layer, geom)
        if crs is None:
            crs = CRS("epsg:4326")
        ext: Geometry | None = None
        # Accumulate extent in native CRS if possible.
        for ds in self.ds_search(layer, times, geom, products):
            if ds.extent:
                if ds.extent.crs != CRS(layer.native_CRS):
                    # Reproject to layer "native" CRS if needed.
                    ds_extent: Geometry = ds.extent.to_crs(layer.native_CRS)
                else:
                    ds_extent = ds.extent
                if ext is None:
                    ext = ds_extent
                else:
                    ext = ext.union(ds_extent)
        if ext is not None and crs != CRS(layer.native_CRS):
            # Reproject to requested CRS if necessary
            return ext.to_crs(crs)
        return ext

    def _prep_geom(self, layer: "OWSNamedLayer", any_geom: Geometry | None) -> Geometry | None:
        # Prepare a Geometry for geospatial search
        # Perhaps Core can be updated so this is not needed?
        if any_geom is None:
            # None?  Leave as None
            return None
        if any_geom.geom_type == "Point":
            # Point?  Expand to a polygon covering a single native pixel.
            any_geom = any_geom.to_crs(layer.native_CRS)
            x, y = any_geom.coords[0]
            delta_x, delta_y = layer.cfg_native_resolution
            return polygon(
                (
                    (x, y),
                    (x + delta_x, y),
                    (x + delta_x, y + delta_y),
                    (x, y + delta_y),
                    (x, y),
                ),
                crs=layer.native_CRS
            )
        elif any_geom.geom_type in ("MultiPoint", "LineString", "MultiLineString"):
            # Not a point, but not a polygon or multipolygon?  Expand to polygon by taking convex hull
            return any_geom.convex_hull
        else:
            # Return polygons and multipolygons as is.
            return any_geom


class OWSAbstractIndexDriver(ABC):
    @classmethod
    @abstractmethod
    def ows_index_class(cls) -> type[OWSAbstractIndex]:
        ...

    @classmethod
    @abstractmethod
    def ows_index(cls) -> OWSAbstractIndex:
        ...


def ows_index(odc: Datacube | AbstractIndex) -> OWSAbstractIndex:
    if isinstance(odc, AbstractIndex):
        index = odc
    else:
        index = odc.index
    env = index.environment
    from datacube_ows.index.driver import ows_index_driver_by_name
    if env.index_driver in ('default', 'legacy'):
        idx_drv_name = "postgres"
    else:
        idx_drv_name = env.index_driver
    ows_index_driver = ows_index_driver_by_name(idx_drv_name)
    if ows_index_driver is None:
        raise ConfigException(f"ODC Environment {env._name} uses ODC index driver {env.index_driver} which is "
                              f"not (yet) supported by OWS.")
    return ows_index_driver.ows_index()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0


from typing import Iterable, Tuple, Union

from xarray import Dataset


def lccs_map_delta(data: Dataset) -> Dataset:
    data1, data2 = (data.sel(time=dt) for dt in data.coords["time"].values)

    equality_mask = data1["level4"] != data2["level4"]
    data1["level4"] = data2["level4"].where(equality_mask, other=255)
    return data1




from datacube import Datacube
from yaml import dump

dc = Datacube()

i = 1
for ds in dc.index.datasets.search(product="s2_l2a"):
    filename = "s2_l2a_ds_%02d.yaml" % i
    with open(filename, "w") as fp:
        fp.write(dump(ds.metadata_doc, default_flow_style=False))
    print(filename, ds.uri)

    i = i + 1



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0


legend_idx_0_1_5ticks = {"begin": "0.0", "end": "1.0", "ticks_every": "0.2"}

legend_idx_percentage_by_10 = {
    "begin": "0.0",
    "end": "1.0",
    "ticks_every": 0.1,
    "units": "%",
    "tick_labels": {
        "0.0": {"label": "0"},
        "0.1": {"label": "10"},
        "0.2": {"label": "20"},
        "0.3": {"label": "30"},
        "0.4": {"label": "40"},
        "0.5": {"label": "50"},
        "0.6": {"label": "60"},
        "0.7": {"label": "70"},
        "0.8": {"label": "80"},
        "0.9": {"label": "90"},
        "1.0": {"label": "100"},
    },
}

legend_idx_percentage_by_20 = {
    "begin": "0.0",
    "end": "1.0",
    "decimal_places": 1,
    "ticks_every": "0.2",
    "units": "%",
    "tick_labels": {
        "0.0": {"label": "0"},
        "0.2": {"label": "20"},
        "0.4": {"label": "40"},
        "0.6": {"label": "60"},
        "0.8": {"label": "80"},
        "1.0": {"label": "100"},
    },
}

legend_idx_percentage_by_25 = {
    "units": "%",
    "decimal_places": 2,
    "begin": "0.00",
    "end": "1.00",
    "ticks_every": 0.25,
    "tick_labels": {
        "0.00": {"label": "0"},
        "0.25": {"label": "25"},
        "0.50": {"label": "50"},
        "0.75": {"label": "75"},
        "1.00": {"label": "100"},
    },
}

legend_idx_twentyplus_3ticks = {
    "begin": 0,
    "end": 20,
    "decimal_places": 0,
    "ticks_every": 10,
    "tick_labels": {"20": {"prefix": ">"}},
}

legend_idx_thirtyplus_4ticks = {
    "begin": 0,
    "end": 30,
    "decimal_places": 0,
    "ticks_every": 10,
    "tick_labels": {"30": {"prefix": ">"}},
    "strip_location": [0.05, 0.5, 0.89, 0.15],
}

legend_idx_0_100_as_0_1_5ticks = {
    "begin": 0,
    "end": 100,
    "units": "unitless",
    "ticks_every": 20,
    "tick_labels": {
        "0": {"label": "0.0"},
        "20": {"label": "0.2"},
        "40": {"label": "0.4"},
        "60": {"label": "0.6"},
        "80": {"label": "0.8"},
        "100": {"label": "1.0"},
    },
}

legend_idx_0_100_pixel_fc_25ticks = {
    "begin": 0,
    "end": 100,
    "units": "% / pixel",
    "ticks_every": 25,
    "title": "Percentage of Pixel that is Green Vegetation",
    "rcParams": {"font.size": 9},
}

legend_idx_0_100_pixel_fc_ngv_25ticks = {
    "begin": 0,
    "end": 100,
    "units": "% / pixel",
    "ticks_every": 25,
    "title": "Percentage of Pixel that is Green Vegetation",
    "rcParams": {"font.size": 9},
}

legend_idx_0_100_pixel_fc_bs_25ticks = {
    "begin": 0,
    "end": 100,
    "ticks_every": 25,
    "units": "% / pixel",
    "title": "Percentage of Pixel that is Bare Soil",
    "rcParams": {"font.size": 9},
}




# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0


#
#  Note this is NOT the configuration file!
#
#  This is a Python module containing the classes and functions used to load and parse configuration files.
#
#  Refer to the documentation for information on how to configure datacube_ows.
#
import datetime
import json
import logging
import math
import os
from collections.abc import Mapping
from enum import Enum
from importlib import import_module
from typing import Any, Optional, Union, cast
from typing_extensions import override
from collections.abc import Iterable

import numpy
from babel.messages.catalog import Catalog
from babel.messages.pofile import read_po
from datacube import Datacube
from datacube.api.query import GroupBy
from datacube.model import Measurement, Product
from datacube.cfg import ODCConfig, ODCEnvironment
from odc.geo import CRS
from odc.geo.geobox import GeoBox
from ows import Version
from slugify import slugify

from datacube_ows.config_utils import (CFG_DICT, RAW_CFG, ConfigException,
                                       F, FlagProductBands, FunctionWrapper,
                                       ODCInitException, OWSConfigEntry,
                                       OWSEntryNotFound, OWSExtensibleConfigEntry,
                                       OWSFlagBand, OWSMetadataConfig,
                                       cfg_expand, get_file_loc,
                                       import_python_obj, load_json_obj)
from datacube_ows.index.api import OWSAbstractIndex, ows_index
from datacube_ows.ogc_utils import create_geobox
from datacube_ows.resource_limits import (OWSResourceManagementRules,
                                          parse_cache_age)
from datacube_ows.styles import StyleDef
from datacube_ows.tile_matrix_sets import TileMatrixSet
from datacube_ows.time_utils import local_solar_date_range
from datacube_ows.utils import (group_by_begin_datetime, group_by_mosaic,
                                group_by_solar)

TYPE_CHECKING = False
if TYPE_CHECKING:
    from datacube_ows.product_ranges import LayerExtent

_LOG = logging.getLogger(__name__)


def read_config(path: str | None = None) -> CFG_DICT:
    cwd = None
    if path:
        cfg_env: str | None = path
    else:
        cfg_env = os.environ.get("DATACUBE_OWS_CFG")
    if not cfg_env:
        from datacube_ows.ows_cfg import ows_cfg as cfg
    elif "/" in cfg_env or cfg_env.endswith(".json"):
        cfg = load_json_obj(cfg_env)
        cwd = get_file_loc(cfg_env)
    elif "." in cfg_env:
        cfg = import_python_obj(cfg_env)
    elif cfg_env.startswith("{"):
        cfg = json.loads(cfg_env)
        abs_path = os.path.abspath(cfg_env)
        cwd = os.path.dirname(abs_path)
    else:
        mod = import_module("datacube_ows.ows_cfg")
        cfg = getattr(mod, cfg_env)
    return cfg_expand(cfg, cwd=cwd)


class BandIndex(OWSMetadataConfig):
    METADATA_DEFAULT_BANDS = True
    METADATA_TITLE = False
    METADATA_ABSTRACT = False

    def __init__(self, layer: "OWSNamedLayer", band_cfg: CFG_DICT) -> None:
        if band_cfg is None:
            band_cfg = {}
        super().__init__(band_cfg)
        self.band_cfg = cast(dict[str, list[str]], band_cfg)
        self.layer = layer
        self.layer_name = layer.name
        self.parse_metadata(band_cfg)
        self._idx: dict[str, str] = {}
        self.add_aliases(self.band_cfg)
        self.declare_unready("_nodata_vals")
        self.declare_unready("measurements")
        self.declare_unready("_dtypes")

    @override
    def global_config(self) -> "OWSConfig":
        return self.layer.global_config()

    @override
    def get_obj_label(self) -> str:
        return self.layer.get_obj_label() + ".bands"

    def add_aliases(self, cfg: dict[str, list[str]]) -> None:
        for b, aliases in cfg.items():
            if b in self._idx:
                raise ConfigException(f"Duplicate band name/alias: {b} in layer {self.layer_name}")
            self._idx[b] = b
            for a in aliases:
                if a != b and a in self._idx:
                    raise ConfigException(f"Duplicate band name/alias: {a} in layer {self.layer_name}")
                self._idx[a] = b

    @override
    def make_ready(self, *args: Any, **kwargs: Any) -> None:
        def floatify_nans(inp: float | int | str) -> float | int:
            if isinstance(inp, str) and inp == "nan":
                return float(inp)
            elif isinstance(inp, str):
                raise ValueError("Invalid nodata value: {inp}")
            else:
                return inp
        default_to_all = not bool(self._raw_cfg)
        # pylint: disable=attribute-defined-outside-init
        self.measurements: dict[str, Measurement] = {}
        self._nodata_vals: dict[str, int | float] = {}
        self._dtypes: dict[str, numpy.dtype] = {}
        first_product = True
        for product in self.layer.products:
            if first_product and default_to_all:
                native_bands = self.layer.dc.list_measurements().loc[product.name]
                for b in native_bands.index:
                    self.band_cfg[b] = [b]
                self.add_aliases(self.band_cfg)
            try:
                prod_measurements = cast(
                    dict[str, Measurement], product.lookup_measurements(list(self.band_cfg.keys()))
                )
                if first_product:
                    self.measurements = prod_measurements
                    self._nodata_vals = {name: floatify_nans(model.nodata) for name, model in self.measurements.items()}
                    self._dtypes = {name: numpy.dtype(model.dtype) for name, model in self.measurements.items()}
                else:
                    for k in prod_measurements:
                        nodata = self._nodata_vals[k]
                        if ((numpy.isnan(nodata) and not numpy.isnan(floatify_nans(prod_measurements[k].nodata)))
                                or (not numpy.isnan(nodata) and prod_measurements[k].nodata != nodata)):
                            raise ConfigException(
                                f"Nodata value mismatch between products for band {k} in multiproduct layer {self.layer_name}")
                        if prod_measurements[k].dtype != self._dtypes[k]:
                            raise ConfigException(
                                f"Data type mismatch between products for band {k} in multiproduct layer {self.layer_name}")
            except KeyError as e:
                raise ConfigException(f"Product {product.name} in layer {self.layer_name} is missing band {e}")
            first_product = False
        super().make_ready(*args, **kwargs)

    def band(self, name_alias: str) -> str:
        if name_alias in self._idx:
            return self._idx[name_alias]
        raise ConfigException(f"Unknown band name/alias: {name_alias} in layer {self.layer_name}")

    def locale_band(self, name_alias: str) -> str:
        try:
            return self.band(name_alias)
        except ConfigException:
            pass
        for b in self.band_cfg.keys():
            if name_alias == self.band_label(b):
                return b
        raise ConfigException(f"Unknown band: {name_alias} in layer {self.layer_name}")

    def band_label(self, name_alias) -> str:
        canonical_name = self.band(name_alias)
        return cast(str, self.read_local_metadata(canonical_name))

    def nodata_val(self, name_alias: str) -> float | int:
        name = self.band(name_alias)
        return self._nodata_vals[name]

    def dtype_val(self, name_alias: str) -> numpy.dtype:
        name = self.band(name_alias)
        return self._dtypes[name]

    def dtype_size(self, name_alias: str) -> int:
        return self.dtype_val(name_alias).itemsize

    def band_labels(self) -> list[str]:
        return [self.band_label(b) for b in self.band_cfg]

    def band_nodata_vals(self) -> list[int | float]:
        return [self.nodata_val(b) for b in self.band_cfg if b in self.band_cfg]


class AttributionCfg(OWSConfigEntry):
    def __init__(self, cfg: CFG_DICT, owner: Union["OWSConfig", "OWSLayer"]) -> None:
        super().__init__(cfg)
        self.owner = owner
        self.url = cast(str | None, cfg.get("url"))
        logo = cast(dict[str, str] | None, cfg.get("logo"))
        if not self.title and not self.url and not logo:
            raise ConfigException("At least one of title, url and logo is required in an attribution definition")
        if not logo:
            self.logo_width: int | None = None
            self.logo_height: int | None = None
            self.logo_url: str | None = None
            self.logo_fmt: str | None = None
        else:
            self.logo_width = None if logo.get("width") is None else int(logo["width"])
            self.logo_height = None if logo.get("height") is None else int(logo["height"])
            self.logo_url = logo.get("url")
            self.logo_fmt = logo.get("format")
            if not self.logo_url or not self.logo_fmt:
                raise ConfigException("url and format must both be specified in an attribution logo.")

    @property
    def title(self) -> str:
        return self.owner.attribution_title  # ????

    @classmethod
    def parse(cls, cfg: CFG_DICT | None, owner: Union["OWSConfig", "OWSLayer"]) -> Optional["AttributionCfg"]:
        if not cfg:
            return None
        else:
            return cls(cfg, owner)


class SuppURL(OWSConfigEntry):
    @classmethod
    def parse_list(cls, cfg: list[dict[str, str]] | None) -> list["SuppURL"]:
        if not cfg:
            return []
        return [cls(u) for u in cfg]

    def __init__(self, cfg: dict[str, str]) -> None:
        super().__init__(cast(RAW_CFG, cfg))
        self.url = cfg["url"]
        self.format = cfg["format"]


class OWSLayer(OWSMetadataConfig):
    METADATA_KEYWORDS = True
    METADATA_ATTRIBUTION = True

    named = False
    def __init__(self, cfg: CFG_DICT, object_label: str, parent_layer: Optional["OWSLayer"]=None, **kwargs) -> None:
        super().__init__(cfg, **kwargs)
        self.object_label = object_label
        self.global_cfg: OWSConfig = kwargs["global_cfg"]
        self.parent_layer = parent_layer
        self._cached_local_env: ODCEnvironment | None = None
        self._cached_dc: Datacube | None = None
        self.parse_metadata(cfg)
        # Do we have a local ODC environment override?
        local_env = cfg.get("env")
        if local_env is not None:
            self._local_env = ODCConfig.get_environment(env=str(local_env))
        # Inherit or override attribution
        if "attribution" in cfg:
            self.attribution = AttributionCfg.parse(
                cast(CFG_DICT | None, cfg.get("attribution")),
                self
            )
        elif parent_layer:
            self.attribution = cast(OWSLayer, self.parent_layer).attribution
        else:
            self.attribution = self.global_cfg.attribution

    @property
    def local_env(self) -> ODCEnvironment:
        # If we have cached inherited environment, use it.
        if self._cached_local_env:
            return self._cached_local_env

        # If we have our own custom environment, use it.
        if hasattr(self, "_local_env") and self._local_env is not None:
            return self._local_env

        # If we have a parent layer, then it's their problem.
        if self.parent_layer:
            # remember, so we don't have to ask our parent layer again.  As a layer we must learn to adult.
            self._cached_local_env = self.parent_layer.local_env
        else:
            # If we have no parent layer, then we have to ask the global government.
            # and remember, so we don't have to deal with the global government again.
            self._cached_local_env = self.global_cfg.default_env

        return self._cached_local_env

    @property
    def dc(self) -> Datacube:
        # If we have cached inherited datacube, use it.
        if self._cached_dc:
            return self._cached_dc

        # If we have our own custom dc, use it.
        if hasattr(self, "_dc"):
            # pylint: disable=access-member-before-definition
            return self._dc  # type: ignore[has-type]

        # If we have our own custom environment, try to make a Datacube from it:
        if hasattr(self, "_local_env") and self._local_env is not None:
            try:
                self._dc = Datacube(env=self._local_env, app=self.global_cfg.odc_app)
                return self._dc
            except Exception as e:
                raise ODCInitException(str(e)) from None

        # If we have a parent layer, then it's their problem.
        if self.parent_layer:
            # remember, so we don't have to ask our parent layer again.  As a layer we must learn to adult.
            self._cached_dc = self.parent_layer.dc
        else:
            # If we have no parent layer, then we have to ask the global government.
            # and remember, so we don't have to deal with the global government again.
            self._cached_dc = self.global_cfg.dc

        return self._cached_dc

    def ows_index(self) -> OWSAbstractIndex:
        return ows_index(self.dc)

    @override
    def global_config(self) -> "OWSConfig":
        return self.global_cfg

    @override
    def can_inherit_from(self) -> Union["OWSConfig", "OWSLayer"]:
        if self.parent_layer:
            return self.parent_layer
        else:
            return self.global_cfg

    @override
    def get_obj_label(self) -> str:
        return self.object_label

    def layer_count(self) -> int:
        return 0

    def unready_layer_count(self) -> int:
        return 0

    @override
    def __str__(self) -> str:
        return "OWSLayer Config: %s" % self.title


class OWSFolder(OWSLayer):
    def __init__(self, cfg: CFG_DICT, global_cfg: "OWSConfig",
                 parent_layer: Optional["OWSFolder"] = None,
                 sibling: int = 0, **kwargs) -> None:
        if "label" in cfg:
            obj_lbl = f"folder.{cfg['label']}"
        elif parent_layer:
            obj_lbl = f"{parent_layer.object_label}.{sibling}"
        else:
            obj_lbl = f"folder.{sibling}"
        if obj_lbl in global_cfg.folder_index:
            raise ConfigException(f"Duplicate folder label: {obj_lbl}")
        super().__init__(cfg, parent_layer=parent_layer, object_label=obj_lbl, global_cfg=global_cfg, **kwargs)
        self.slug_name = slugify(self.title, separator="_")
        self.unready_layers: list[OWSLayer] = []
        self.child_layers: list[OWSLayer] = []
        if "layers" not in cfg:
            raise ConfigException("No layers section in folder layer %s" % self.title)
        child = 0
        for lyr_cfg in cast(list[RAW_CFG], cfg["layers"]):
            if isinstance(lyr_cfg, dict):
                try:
                    lyr = parse_ows_layer(lyr_cfg, global_cfg=global_cfg, parent_layer=self, sibling=child)
                    self.unready_layers.append(lyr)
                except ConfigException as e:
                    _LOG.error("Could not parse layer (%s): %s",
                               lyr_cfg.get("name", lyr_cfg.get("title", "??")),
                               str(e))
                child += 1
            else:
                _LOG.error("Non-dictionary where dictionary expected - check for trailing comma? %s...", repr(lyr_cfg)[0:50])
        global_cfg.folder_index[obj_lbl] = self

    @override
    def unready_layer_count(self) -> int:
        return sum([l.layer_count() for l in self.unready_layers])

    @override
    def layer_count(self) -> int:
        return sum([l.layer_count() for l in self.child_layers])

    @override
    def make_ready(self, *args, **kwargs) -> None:
        still_unready = []
        for lyr in self.unready_layers:
            try:
                lyr.make_ready(*args, **kwargs)
                self.child_layers.append(lyr)
            except ConfigException as e:
                _LOG.error("Could not load layer %s: %s", lyr.title, str(e))
                still_unready.append(lyr)
        self.unready_layers = still_unready
        super().make_ready(*args, **kwargs)

    @override
    def __repr__(self) -> str:
        return f"OWS Folder <{self.title}>"


class TimeRes(Enum):
    SUBDAY = "subday"
    SOLAR = "solar"
    SUMMARY = "summary"

    def is_subday(self) -> bool:
        return self == self.SUBDAY

    def is_solar(self) -> bool:
        return self  == self.SOLAR

    def is_summary(self) -> bool:
        return not self.is_solar() and not self.is_subday()

    def allow_mosaic(self) -> bool:
        return not self.is_subday()

    @classmethod
    def parse(cls, cfg: str | None) -> Optional["TimeRes"]:
        if cfg is None:
            cfg = "solar"
        elif cfg == "raw":
            _LOG.warning("The 'raw' time resolution type is deprecated.  Please use 'solar'.")
            cfg = "solar"
        elif cfg in ("day", "month", "year"):
            _LOG.warning("The '%s' time resolution type is deprecated.  Please use 'summary'.", cfg)
            cfg = "summary"
        try:
            return cls(cfg)
        except ValueError:
            return None

    def search_times(self,
                     t: datetime.datetime,
                     geobox: GeoBox | None = None) -> datetime.datetime | tuple[datetime.datetime, datetime.datetime]:
        if self.is_solar():
            if geobox is None:
                raise ValueError("Solar time resolution search_times requires a geobox.")
            times: datetime.datetime | tuple[datetime.datetime, datetime.datetime] = local_solar_date_range(geobox, t)
        elif self.is_subday():
            # For subday products, return a single start datetime instead of a range.
            # mv_index will expand this to a one-second search range.
            # This prevents users from having to always use the full ISO timestamp in queries.
            times = t
        else:
            # For summary products, return a single start date instead of a range.
            # mv_index will expand this to a one-day search range
            # This allows data with overlapping time periods to be resolved by start date.
            times = t

        return times

    def dataset_groupby(self, product_names: list[str] | None = None, is_mosaic=False) -> GroupBy:
        if self.is_subday():
            return group_by_begin_datetime(product_names, truncate_dates=False)
        elif is_mosaic:
            return group_by_mosaic(product_names)
        elif self.is_solar():
            return group_by_solar(product_names)
        else:
            return group_by_begin_datetime(product_names)

DEF_TIME_LATEST = "latest"
DEF_TIME_EARLIEST = "earliest"


class OWSNamedLayer(OWSExtensibleConfigEntry, OWSLayer):
    INDEX_KEYS = ["layer"]
    named = True
    multi_product: bool = False

    def __init__(self, cfg: CFG_DICT, global_cfg: "OWSConfig", parent_layer: OWSFolder | None = None, **kwargs) -> None:
        name = cast(str, cfg["name"])
        super().__init__(cfg, object_label=f"layer.{name}", global_cfg=global_cfg, parent_layer=parent_layer,
                         keyvals={"layer": name},
                         **kwargs)
        self.name = name
        cfg = cast(CFG_DICT, self._raw_cfg)
        self.hide = False
        try:
            self.parse_product_names(cfg)
            if len(self.low_res_product_names) not in (0, len(self.product_names)):
                raise ConfigException(f"Lengths of product_names and low_res_product_names do not match in layer {self.name}")
            for prod_name in self.product_names:
                if "__" in prod_name:
                    # I think this was for subproducts which are currently broken
                    raise ConfigException("Product names cannot contain a double underscore '__'.")
        except IndexError:
            raise ConfigException(f"No products declared in layer {self.name}")
        except KeyError as e:
            raise ConfigException("Required product names entry (%s) missing in named layer %s" % (
                str(e),
                self.name
            ))
        self.declare_unready("products")
        self.declare_unready("low_res_products")
        self.declare_unready("product")
        self.declare_unready("definition")

        if global_cfg.user_band_math_extension:
            self.user_band_math = bool(cfg.get("user_band_math", False))
        else:
            self.user_band_math = False
        tr = TimeRes.parse(cast(str | None, cfg.get("time_resolution")))
        if not tr:
            raise ConfigException(f"Invalid time resolution value {cfg['time_resolution']} in named layer {self.name}")
        else:
            self.time_resolution: TimeRes = tr
        self.mosaic_date_func: FunctionWrapper | None = None
        if "mosaic_date_func" in cfg:
            self.mosaic_date_func = FunctionWrapper(self, cast(CFG_DICT, cfg["mosaic_date_func"]))
        if self.mosaic_date_func and not self.time_resolution.allow_mosaic():
            raise ConfigException(f"Mosaic date function not supported for {self.time_resolution} time resolution.")
        dtr: str = cast(str, cfg.get("default_time", DEF_TIME_LATEST))
        if dtr in (DEF_TIME_LATEST, DEF_TIME_EARLIEST):
            self.default_time_rule: str | datetime.datetime | datetime.date = dtr
        else:
            try:
                if self.time_resolution.is_subday():
                    self.default_time_rule = datetime.datetime.fromisoformat(dtr)
                else:
                    self.default_time_rule = datetime.date.fromisoformat(dtr)
            except ValueError:
                raise ConfigException(
                    f"Invalid default_time value in named layer {self.name} ({dtr})"
                )
        self.time_axis = cast(CFG_DICT | None, cfg.get("time_axis"))
        if self.time_axis:
            if self.time_resolution.is_subday():
                raise ConfigException("Regular time axis is not supported for sub-day time resolutions")
            self.regular_time_axis = True
            if "time_interval" not in self.time_axis:
                raise ConfigException("No time_interval supplied in time_axis")
            time_axis_interval = self.time_axis["time_interval"]
            if isinstance(time_axis_interval, int):
                self.time_axis_interval: int = time_axis_interval
            else:
                raise ConfigException("time_interval must be an integer")
            if self.time_axis_interval <= 0:
                raise ConfigException("time_interval must be greater than zero")
            time_axis_start = cast(str | None, self.time_axis.get("start_date"))
            time_axis_end = cast(str | None, self.time_axis.get("end_date"))
            if time_axis_start is None:
                self.time_axis_start: datetime.date | None = None
            else:
                try:
                    self.time_axis_start = datetime.date.fromisoformat(time_axis_start)
                except ValueError:
                    raise ConfigException("time_axis start_date is not a valid ISO format date string")
            if time_axis_end is None:
                self.time_axis_end: datetime.date | None = None
            else:
                try:
                    self.time_axis_end = datetime.date.fromisoformat(time_axis_end)
                except ValueError:
                    raise ConfigException("time_axis end_date is not a valid ISO format date string")
            if (self.time_axis_end is not None
                    and self.time_axis_start is not None
                    and self.time_axis_end < self.time_axis_start):
                raise ConfigException("time_axis end_date must be greater than or equal to the start_date if both are provided")
        else:
            self.regular_time_axis = False
            self.time_axis_interval = 0
            self.time_axis_start = None
            self.time_axis_end = None

        self.dynamic = cfg.get("dynamic", False)

        self.declare_unready("default_time")
        self.declare_unready("_ranges")
        self.declare_unready("bboxes")
        self.band_idx: BandIndex = BandIndex(self, cast(CFG_DICT, cfg.get("bands")))
        self.cfg_native_resolution = cfg.get("native_resolution")
        self.cfg_native_crs = cfg.get("native_crs")
        self.declare_unready("resolution_x")
        self.declare_unready("resolution_y")
        self.resource_limits = OWSResourceManagementRules(self.global_cfg,
                                                          cast(CFG_DICT, cfg.get("resource_limits", {})),
                                                          f"Layer {self.name}")
        try:
            self.parse_flags(cast(CFG_DICT, cfg.get("flags", {})))
            self.declare_unready("all_flag_band_names")
        except KeyError as e:
            raise ConfigException(f"Missing required config ({str(e)}) in flags section for layer {self.name}")
        try:
            self.parse_image_processing(cast(CFG_DICT, cfg["image_processing"]))
        except KeyError as e:
            raise ConfigException(f"Missing required config ({str(e)}) in image processing section for layer {self.name}")
        self.identifiers = cast(dict[str, str], cfg.get("identifiers", {}))
        for auth in self.identifiers.keys():
            if auth not in self.global_cfg.authorities:
                raise ConfigException(f"Identifier with non-declared authority: {auth} in layer {self.name}")
        self.parse_urls(cast(CFG_DICT, cfg.get("urls", {})))
        self.parse_feature_info(cast(CFG_DICT, cfg.get("feature_info", {})))
        self.feature_info_include_utc_dates = cfg.get("feature_info_url_dates", False)
        if "patch_url_function" in cfg:
            self.patch_url: FunctionWrapper | None = FunctionWrapper(self, cast(CFG_DICT, cfg["patch_url_function"]))
        else:
            self.patch_url = None
        try:
            self.parse_styling(cast(CFG_DICT, cfg["styling"]))
        except KeyError as e:
            raise ConfigException(f"Missing required config item {e} in styling section for layer {self.name}")

        if self.global_cfg.wcs:
            try:
                self.parse_wcs(cast(CFG_DICT | bool, cfg.get("wcs", {})))
            except KeyError as e:
                raise ConfigException(f"Missing required config item {e} in wcs section for layer {self.name}")

#       Sub-products have been broken for some time.
#        sub_prod_cfg = cfg.get("sub_products", {})
#        self.sub_product_label = sub_prod_cfg.get("label")
#        if "extractor" in sub_prod_cfg:
#            self.sub_product_extractor = FunctionWrapper(self, sub_prod_cfg["extractor"])
#        else:
#            self.sub_product_extractor = None
        # And finally, add to the global product index.
        existing = self.global_cfg.layer_index.get(self.name)
        if existing and existing != self:
            raise ConfigException(f"Duplicate layer name: {self.name}")
        self.global_cfg.layer_index[self.name] = self

    def time_axis_representation(self) -> str:
        if self.regular_time_axis:
            start, end = self.time_range(self.ranges)
            return f"{start.isoformat()}/{end.isoformat()}/P{self.time_axis_interval}D"
        return ""

    # pylint: disable=attribute-defined-outside-init
    @override
    def make_ready(self, *args: Any, **kwargs: Any) -> None:
        self.products: list[Product] = []
        self.low_res_products: list[Product] = []
        for i, prod_name in enumerate(self.product_names):
            if self.low_res_product_names:
                low_res_prod_name = self.low_res_product_names[i]
            else:
                low_res_prod_name = None
            product = self.dc.index.products.get_by_name(prod_name)
            if not product:
                raise ConfigException(f"Could not find product {prod_name} in datacube for layer {self.name}")
            self.products.append(product)
            if low_res_prod_name:
                product = self.dc.index.products.get_by_name(low_res_prod_name)
                if not product:
                    raise ConfigException(
                        f"Could not find product {low_res_prod_name} in datacube for layer {self.name}"
                    )
                self.low_res_products.append(product)
        self.product = self.products[0]
        self.definition = self.product.definition
        self.force_range_update()
        self.band_idx.make_ready()
        self.resource_limits.make_ready()
        self.all_flag_band_names: set[str] = set()
        for fb in self.flag_bands.values():
            fb.make_ready()
            if fb.pq_band in self.all_flag_band_names:
                raise ConfigException(f"Duplicate flag band name: {fb.pq_band}")
            self.all_flag_band_names.add(fb.pq_band)
        self.ready_image_processing()
        self.ready_native_specs()
        if self.global_cfg.wcs:
            self.ready_wcs()
        for style in self.styles:
            style.make_ready(*args, **kwargs)
        for fpb in self.allflag_productbands:
            fpb.make_ready(*args, **kwargs)
        if not self.multi_product:
            self.global_cfg.native_product_index[self.product_name] = self

        if not self.hide:
            super().make_ready(*args, **kwargs)

    # pylint: disable=attribute-defined-outside-init
    def parse_image_processing(self, cfg: CFG_DICT) -> None:
        emf_cfg = cfg["extent_mask_func"]
        if isinstance(emf_cfg, dict) or isinstance(emf_cfg, str):
            self.extent_mask_func = [FunctionWrapper(self, emf_cfg)]
        else:
            self.extent_mask_func = [
                FunctionWrapper(self, emf) for emf in cast(list[CFG_DICT | str], emf_cfg)
            ]
        self.raw_afb = cfg.get("always_fetch_bands", [])
        self.declare_unready("always_fetch_bands")
        self.solar_correction = bool(cfg.get("apply_solar_corrections", False))
        self.data_manual_merge = bool(cfg.get("manual_merge", False))
        if self.solar_correction and not self.data_manual_merge:
            raise ConfigException("Solar correction requires manual_merge.")
        if self.data_manual_merge and not self.solar_correction and not self.multi_product:
            _LOG.warning("Manual merge is only recommended where solar correction is required and for multi-product layers.")

        if cfg.get("fuse_func"):
            self.fuse_func: FunctionWrapper | None = FunctionWrapper(
                self,
                cast(str | CFG_DICT, cfg["fuse_func"])
            )
        else:
            self.fuse_func = None

    # pylint: disable=attribute-defined-outside-init
    def ready_image_processing(self) -> None:
        self.always_fetch_bands = list([self.band_idx.band(b) for b in cast(list[str], self.raw_afb)])

    # pylint: disable=attribute-defined-outside-init
    def parse_feature_info(self, cfg: CFG_DICT) -> None:
        self.feature_info_include_utc_dates = bool(cfg.get("include_utc_dates", False))
        custom = cast(dict[str, CFG_DICT | str], cfg.get("include_custom", {}))
        self.legacy_feature_info_custom_includes = {k: FunctionWrapper(self, v) for k, v in custom.items()}
        if self.legacy_feature_info_custom_includes:
            _LOG.warning("In layer %s: The 'include_custom' directive is deprecated and will be removed in version 1.9. "
                         "Please refer to the documentation for information on how to migrate your configuration "
                         "to the new 'custom_includes' directive.", self.name)
        custom = cast(dict[str, CFG_DICT | str | F], cfg.get("custom_includes", {}))
        self.feature_info_custom_includes = {k: FunctionWrapper(self, v) for k, v in custom.items()}

    # pylint: disable=attribute-defined-outside-init
    def parse_flags(self, cfg: CFG_DICT) -> None:
        self.flag_bands = {}
        if cfg:
            if isinstance(cfg, dict):
                fb = OWSFlagBand(cfg, self)
                self.flag_bands[fb.pq_band] = fb
                _LOG.warning("Single flag bands not in a list is deprecated. Please refer to the documentation for the new format (layer %s)", self.name)
            else:
                for fb_cfg in cfg:
                    fb = OWSFlagBand(fb_cfg, self)
                    self.flag_bands[fb.pq_band] = fb
        pq_names_to_lowres_names: dict[list[str], list[str]] = {}
        for fb in self.flag_bands.values():
            pns = fb.pq_names
            lrpns = fb.pq_low_res_names
            if pns in pq_names_to_lowres_names and pq_names_to_lowres_names[pns] != lrpns:
                raise ConfigException(f"Product name mismatch in flags section for layer {self.name}: product_names {pns} has multiple distinct low-res product names")
            pq_names_to_lowres_names[pns] = lrpns
        # pylint: disable=dict-values-not-iterating
        self.allflag_productbands = FlagProductBands.build_list_from_flagbands(self.flag_bands.values(), self)

    # pylint: disable=attribute-defined-outside-init
    def parse_urls(self, cfg: CFG_DICT) -> None:
        self.feature_list_urls = SuppURL.parse_list(cast(list[dict[str, str]], cfg.get("features", [])))
        self.data_urls = SuppURL.parse_list(cast(list[dict[str, str]], cfg.get("data", [])))

    # pylint: disable=attribute-defined-outside-init
    def parse_styling(self, cfg: CFG_DICT) -> None:
        self.styles = []
        self.style_index = {}
        for scfg in cast(list[CFG_DICT], cfg["styles"]):
            style = StyleDef(self, scfg)
            self.styles.append(style)
            self.style_index[style.name] = style
        if "default_style" in cfg:
            if cfg["default_style"] not in self.style_index:
                raise ConfigException(f"Default style {cfg['default_style']} is not in the 'styles' for layer {self.name}")
            self.default_style = self.style_index[cast(str, cfg["default_style"])]
        else:
            self.default_style = self.styles[0]

    # pylint: disable=attribute-defined-outside-init
    def parse_wcs(self, cfg: CFG_DICT | bool) -> None:
        if cfg == False or not self.global_cfg.wcs:
            self.wcs = False
        else:
            self.wcs = not cast(CFG_DICT, cfg).get("disable", False)
        if not self.wcs:
            return
        assert isinstance(cfg, dict)
        if "native_resolution" in cfg:
            if not self.cfg_native_resolution:
                _LOG.warning(
                    "Specifying native_resolution in wcs section of layer %s is now deprecated, please move to "
                    "main layer section if required.", self.name)
                self.cfg_native_resolution = cfg.get("native_resolution")
            else:
                _LOG.warning(
                    "Native_resolution in wcs section of layer %s ignored in favour of value in "
                    "main layer section.", self.name)

        # Native CRS
        if "native_crs" in cfg:
            if not self.cfg_native_crs:
                _LOG.warning("Specifying native_crs in wcs section of layer %s is now deprecated, pleas move to "
                             "main layer section if required", self.name)
                self.cfg_native_crs = cfg["native_crs"]
            else:
                _LOG.warning(
                    "native_crs in wcs section of layer %s ignored in favour of value in "
                    "main layer section.", self.name)

        self.declare_unready("native_CRS")
        self.declare_unready("native_CRS_def")

        # Rectified Grids
        self.declare_unready("origin_x")
        self.declare_unready("origin_y")
        self.declare_unready("grid_high_x")
        self.declare_unready("grid_high_y")
        self.declare_unready("grids")
        # Band management
        if cfg.get("default_bands"):
            _LOG.warning(
                "wcs section contains a 'default_bands' list.  WCS default_bands list is no longer supported. "
                "Functionally, the default behaviour is now to return all available bands (as mandated by "
                "the WCS2.x spec). "
            )

        # Native format
        if "native_format" in cfg:
            self.native_format = cfg["native_format"]
            if self.native_format not in self.global_cfg.wcs_formats_by_name:
                raise ConfigException(f"WCS native format {self.native_format} for layer {self.name} is not in supported formats list")
        else:
            self.native_format = self.global_cfg.native_wcs_format

    # pylint: disable=attribute-defined-outside-init
    def ready_native_specs(self):
        # Native CRS
        try:
            self.native_CRS = self.product.definition["storage"]["crs"]
            if self.cfg_native_crs == self.native_CRS:
                _LOG.debug(
                    "Native crs for layer %s is specified in ODC metadata and does not need to be specified in configuration",
                    self.name)
            else:
                _LOG.warning("Native crs for layer %s is specified in config as %s - overridden to %s by ODC metadata",
                             self.name, self.cfg_native_crs, self.native_CRS)
        except KeyError:
            self.native_CRS = self.cfg_native_crs

        if not self.native_CRS:
            raise ConfigException(f"No native CRS could be found for layer {self.name}")
        if self.native_CRS not in self.global_cfg.published_CRSs:
            raise ConfigException(
                f"Native CRS for product {self.product_name} in layer {self.name} ({self.native_CRS}) not in published CRSs")
        self.native_CRS_def = self.global_cfg.published_CRSs[self.native_CRS]

        try:
            # Native CRS
            self.resolution_x = self.product.definition["storage"]["resolution"][
                self.native_CRS_def["horizontal_coord"]]
            self.resolution_y = self.product.definition["storage"]["resolution"][self.native_CRS_def["vertical_coord"]]
        except KeyError:
            self.resolution_x = None
            self.resolution_y = None

        if self.resolution_x is None:
            try:
                if self.cfg_native_resolution is None:
                    raise KeyError
                self.resolution_x, self.resolution_y = self.cfg_native_resolution
            except KeyError:
                raise ConfigException(
                    f"No native resolution supplied for layer {self.name} with no product-native resolution defined in ODC."
                )
            except ValueError:
                raise ConfigException(f"Invalid native resolution supplied for layer {self.name}")
            except TypeError:
                raise ConfigException(f"Invalid native resolution supplied for layer {self.name}")
        elif self.cfg_native_resolution:
            config_x, config_y = (float(r) for r in self.cfg_native_resolution)
            if (
                    math.isclose(config_x, float(self.resolution_x), rel_tol=1e-8)
                    and math.isclose(config_y, float(self.resolution_y), rel_tol=1e-8)
            ):
                _LOG.debug(
                    "Native resolution for layer %s is specified in ODC metadata and does not need to be specified in configuration",
                    self.name)
            else:
                _LOG.warning(
                    "Native resolution for layer %s is specified in config as %s - overridden to (%.15f, %.15f) by ODC metadata",
                    self.name, repr(self.cfg_native_resolution), self.resolution_x, self.resolution_y)

    # pylint: disable=attribute-defined-outside-init
    def ready_wcs(self) -> None:
        if self.global_cfg.wcs and self.wcs:
            # Prepare Rectified Grids
            try:
                native_bounding_box = self.bboxes[self.native_CRS]
            except KeyError:
                if not self.global_cfg.called_from_update_ranges:
                    _LOG.warning("Layer: %s No bounding box in ranges for native CRS %s - rerun datacube-ows-update",
                                 self.name,
                                 self.native_CRS)
                self.hide = True
                return
            self.origin_x = native_bounding_box["left"]
            self.origin_y = native_bounding_box["bottom"]


            if abs(native_bounding_box["right"] - native_bounding_box["left"]) < abs(self.resolution_x):
                raise ConfigException(
                    "Native (%s) bounding box on layer %s has left %.8f, right %.8f (diff %d), but horizontal resolution is %.8f"
                    % (
                        self.native_CRS,
                        self.name,
                        native_bounding_box["left"],
                        native_bounding_box["right"],
                        native_bounding_box["right"] - native_bounding_box["left"],
                        self.resolution_x
                    ))
            if abs(native_bounding_box["top"] - native_bounding_box["bottom"]) < abs(self.resolution_y):
                raise ConfigException(
                    "Native (%s) bounding box on layer %s has bottom %f, top %f (diff %d), but vertical resolution is %f"
                    % (
                        self.native_CRS,
                        self.name,
                        native_bounding_box["bottom"],
                        native_bounding_box["top"],
                        native_bounding_box["top"] - native_bounding_box["bottom"],
                        self.resolution_y

                    ))
            self.grid_high_x = abs(int((native_bounding_box["right"] - native_bounding_box["left"]) / self.resolution_x))
            self.grid_high_y = int((native_bounding_box["bottom"] - native_bounding_box["top"]) / self.resolution_y)

            if self.grid_high_x <= 0:
                err_str = f"Grid High x is non-positive on layer {self.name}: native ({self.native_CRS}) extent: {native_bounding_box['left']},{native_bounding_box['right']}: x_res={self.resolution_x}"
                raise ConfigException(err_str)
            if self.grid_high_y <= 0:
                err_str = f"Grid High y is non-positive on layer {self.name}: native ({self.native_CRS}) extent: {native_bounding_box['bottom']},{native_bounding_box['top']}: y_res={self.resolution_y}"
                raise ConfigException(err_str)
            self.grids = {}
            for crs, crs_def in self.global_cfg.published_CRSs.items():
                if crs == self.native_CRS:
                    self.grids[crs] = {
                        "origin": (self.origin_x, self.origin_y),
                        "resolution": (self.resolution_x, self.resolution_y),
                    }
                else:
                    try:
                        bbox = self.bboxes[crs]
                    except KeyError:
                        continue
                    self.grids[crs] = {
                        "origin": (bbox["left"], bbox["bottom"]),
                        "resolution": (
                            (bbox["right"] - bbox["left"]) / self.grid_high_x,
                            (bbox["top"] - bbox["bottom"]) / self.grid_high_y
                        )
                    }

    def parse_product_names(self, cfg: CFG_DICT):
        raise NotImplementedError()

    def parse_pq_names(self, cfg: CFG_DICT):
        raise NotImplementedError()

    def force_range_update(self) -> None:
        self.hide = False
        try:
            self._ranges = self.ows_index().get_ranges(self)
            if self._ranges is None:
                raise Exception("Null product range")
            self.bboxes = self.extract_bboxes()
            if self.default_time_rule == DEF_TIME_EARLIEST:
                self.default_time = cast(datetime.datetime | datetime.date, self._ranges.start_time)
            elif isinstance(self.default_time_rule,
                            datetime.date) and self.default_time_rule in cast(set[datetime.datetime | datetime.date],
                                                                              self._ranges.time_set):
                self.default_time = cast(datetime.datetime | datetime.date, self.default_time_rule)
            elif isinstance(self.default_time_rule, datetime.date):
                _LOG.warning("default_time for named_layer %s is explicit date (%s) that is "
                             " not available for the layer. Using most recent available date instead.",
                                    self.name,
                                    self.default_time_rule.isoformat()
                )
                self.default_time = cast(datetime.datetime | datetime.date, self._ranges.end_time)
            else:
                self.default_time = cast(datetime.datetime | datetime.date, self._ranges.end_time)

        # pylint: disable=broad-except
        except Exception as a:
            if not self.global_cfg.called_from_update_ranges:
                _LOG.warning("get_ranges failed for layer %s: %s", self.name, str(a))
            self.hide = True
            self.bboxes = {}

    def time_range(self,
                   ranges: Optional["LayerExtent"] = None
                   ) -> tuple[datetime.datetime | datetime.date, datetime.datetime | datetime.date]:
        if ranges is None:
            ranges = self.ranges
        if self.regular_time_axis and self.time_axis_start:
            start = self.time_axis_start
        else:
            start = ranges.start_time
        if self.regular_time_axis and self.time_axis_end:
            end = self.time_axis_end
        else:
            end = ranges.end_time
        return start, end

    @property
    def ranges(self) -> "LayerExtent":
        if self.dynamic:
            self.force_range_update()
        assert self._ranges is not None  # For type checker
        return self._ranges

    def extract_bboxes(self) -> dict[str, Any]:
        if self._ranges is None:
            return {}
        bboxes = {}
        for crs_id, bbox in cast(dict[str, dict[str, float]], self._ranges.bboxes).items():
            if crs_id in self.global_cfg.published_CRSs:
                # Assume we've already handled coordinate swapping for
                # Vertical-coord first CRSs.   Top is top, left is left.
                bboxes[crs_id] = {
                    "right": bbox["right"],
                    "left": bbox["left"],
                    "top": bbox["top"],
                    "bottom": bbox["bottom"],
                    "vertical_coord_first": self.global_cfg.published_CRSs[crs_id]["vertical_coord_first"]
                }
        return bboxes

    @override
    def layer_count(self) -> int:
        return 1

    def search_times(self, t, geobox=None) -> datetime.datetime | tuple:
        if not geobox:
            bbox = self.ranges.bboxes[self.native_CRS]
            geobox = create_geobox(
                self.native_CRS,
                bbox["left"], bbox["bottom"], bbox["right"], bbox["top"],
                1, 1
            )
        return self.time_resolution.search_times(t, geobox)

    def dataset_groupby(self) -> GroupBy:
        return self.time_resolution.dataset_groupby(is_mosaic=self.mosaic_date_func is not None)

    @override
    def __str__(self) -> str:
        return "Named OWSLayer: %s" % self.name

    @classmethod
    @override
    def lookup_impl(cls, cfg: "OWSConfig", keyvals: dict[str, str], subs: CFG_DICT | None = None):
        try:
            return cfg.layer_index[keyvals["layer"]]
        except KeyError:
            raise OWSEntryNotFound(f"Layer {keyvals['layer']} not found")

    @override
    def __repr__(self) -> str:
        return f"OWS Layer <{self.name}>"


class OWSProductLayer(OWSNamedLayer):
    multi_product = False

    @override
    def parse_product_names(self, cfg: CFG_DICT) -> None:
        self.product_name = cast(str, cfg["product_name"])
        self.product_names: tuple[str, ...] = (self.product_name,)

        self.low_res_product_name = cast(str, cfg.get("low_res_product_name"))
        if self.low_res_product_name:
            self.low_res_product_names: tuple[str, ...] = (self.low_res_product_name,)
        else:
            self.low_res_product_names = tuple()
        if "product_names" in cfg:
            raise ConfigException(f"'product_names' entry in non-multi-product layer {self.name} - use 'product_name' only")
        if "low_res_product_names" in cfg:
            raise ConfigException(f"'low_res_product_names' entry in non-multi-product layer {self.name} - use 'low_res_product_name' only")

    @override
    def parse_pq_names(self, cfg: CFG_DICT) -> dict[str, tuple | bool]:
        main_product = False
        if "dataset" in cfg:
            raise ConfigException(f"The 'dataset' entry in the flags section is no longer supported.  Please refer to the documentation for the correct format (layer {self.name})")
        if "product" in cfg:
            pq_names: tuple[str, ...] = (cast(str, cfg["product"]),)
        else:
            pq_names = (self.product_name,)
            main_product = (pq_names[0] == self.product_name)

        if "low_res_product" in cfg:
            pq_low_res_names: tuple[str, ...] = (cast(str, cfg.get("low_res_product")),)
        elif main_product:
            pq_low_res_names = self.low_res_product_names
        else:
            pq_low_res_names = pq_names

        if "products" in cfg:
            raise ConfigException(f"'products' entry in flags section of non-multi-product layer {self.name} - use 'product' only")
        if "low_res_products" in cfg:
            raise ConfigException(f"'low_res_products' entry in flags section of non-multi-product layer {self.name}- use 'low_res_product' only")
        return {
            "pq_names": pq_names,
            "pq_low_res_names": pq_low_res_names,
            "main_products": main_product
        }


class OWSMultiProductLayer(OWSNamedLayer):
    multi_product = True

    @override
    def parse_product_names(self, cfg: CFG_DICT) -> None:
        self.product_names = tuple(cast(list[str], cfg["product_names"]))
        self.product_name = self.product_names[0]
        self.low_res_product_names = tuple(cast(list[str], cfg.get("low_res_product_names", [])))
        if self.low_res_product_names:
            self.low_res_product_name: str | None = self.low_res_product_names[0]
        else:
            self.low_res_product_name = None
        if "product_name" in cfg:
            raise ConfigException(f"'product_name' entry in multi-product layer {self.name} - use 'product_names' only")
        if "low_res_product_name" in cfg:
            raise ConfigException(f"'low_res_product_name' entry in multi-product layer {self.name} - use 'low_res_product_names' only")

    @override
    def parse_pq_names(self, cfg: CFG_DICT) -> dict[str, tuple | bool]:
        main_products = False
        if "datasets" in cfg:
            raise ConfigException(f"The 'datasets' entry in the flags section is no longer supported. Please refer to the documentation for the correct format (layer {self.name})")
        if "products" in cfg:
            pq_names = tuple(cast(list[str], cfg["products"]))
            main_products = pq_names == self.product_names
        else:
            main_products = True
            pq_names = self.product_names

        if "low_res_products" in cfg:
            pq_low_res_names = tuple(cast(list[str], cfg["low_res_products"]))
        else:
            pq_low_res_names = self.low_res_product_names
        if "product" in cfg:
            raise ConfigException(f"'product' entry in flags section of multi-product layer {self.name} - use 'products' only")
        if "low_res_product" in cfg:
            raise ConfigException(f"'low_res_product' entry in flags section of multi-product layer {self.name} - use 'low_res_products' only")
        return {
            "pq_names": pq_names,
            "pq_low_res_names": pq_low_res_names,
            "main_products": main_products,
        }

    @override
    def dataset_groupby(self) -> GroupBy:
        return self.time_resolution.dataset_groupby(
            list(self.product_names),
            is_mosaic=self.mosaic_date_func is not None)


def parse_ows_layer(cfg: CFG_DICT,
                    global_cfg: "OWSConfig",
                    parent_layer: OWSFolder | None = None,
                    sibling: int = 0) -> OWSLayer:
    if cfg.get("name", None):
        if cfg.get("multi_product", False):
            return OWSMultiProductLayer(cfg, global_cfg, parent_layer)
        else:
            return OWSProductLayer(cfg, global_cfg, parent_layer)
    else:
        return OWSFolder(cfg, global_cfg, parent_layer=parent_layer, sibling=sibling)


class WCSFormat:
    @staticmethod
    def from_cfg(cfg: dict[str, CFG_DICT]) -> list["WCSFormat"]:
        renderers: list[WCSFormat] = []
        for name, fmt in cfg.items():
            if "renderers" in fmt:
                renderers.append(
                    WCSFormat(
                        name,
                        cast(str, fmt["mime"]),
                        cast(str, fmt["extension"]),
                        cast(dict[int, CFG_DICT], fmt["renderers"]),
                        bool(fmt.get("multi-time", False))
                    )
                )
        return renderers

    def __init__(self, name: str, mime: str, extension: str, renderers: dict[int, CFG_DICT], multi_time: bool) -> None:
        self.name = name
        self.mime = mime
        self.extension = extension
        self.multi_time = multi_time
        self.renderers = {
            int(ver): FunctionWrapper(None, renderer)
            for ver, renderer in renderers.items()
        }
        if 1 not in self.renderers:
            _LOG.warning("No renderer supplied for WCS 1.x for format %s", self.name)
        if 2 not in self.renderers:
            _LOG.warning("Warning: No renderer supplied for WCS 2.x for format %s", self.name)

    def renderer(self, version: str | int | Version) -> FunctionWrapper:
        if isinstance(version, str):
            version = int(version.split(".")[0])
        elif isinstance(version, Version):
            version = version.major
        return self.renderers[version]


class ContactInfo(OWSConfigEntry):
    def __init__(self, cfg: CFG_DICT, global_cfg: "OWSConfig") -> None:
        super().__init__(cfg)
        self.global_cfg = global_cfg
        self.person = cfg.get("person")

        class Address(OWSConfigEntry):
            def __init__(self, cfg: dict[str, str]):
                super().__init__(cast(CFG_DICT, cfg))
                self.type = cfg.get("type")
                self.address = cfg.get("address")
                self.city = cfg.get("city")
                self.state = cfg.get("state")
                self.postcode = cfg.get("postcode")
                self.country = cfg.get("country")

            @classmethod
            def parse(cls, cfg: dict[str, str] | None) -> Optional["Address"]:
                if not cfg:
                    return None
                else:
                    return cls(cfg)

        self.address = Address.parse(cast(dict[str, str] | None, cfg.get("address")))
        self.telephone = cast(str | None, cfg.get("telephone"))
        self.fax = cast(str | None, cfg.get("fax"))
        self.email = cast(str | None, cfg.get("email"))

    @property
    def organisation(self) -> str | None:
        return self.global_cfg.contact_org

    @property
    def position(self) -> str | None:
        return self.global_cfg.contact_position

    @classmethod
    def parse(cls, cfg, global_cfg) -> Optional["ContactInfo"]:
        if cfg:
            return cls(cfg, global_cfg)
        else:
            return None


class OWSConfig(OWSMetadataConfig):
    _instance: Optional["OWSConfig"] = None
    initialised = False

    def __new__(cls, *args, **kwargs) -> "OWSConfig":
        if not cls._instance or kwargs.get("refresh"):
            cls._instance = super().__new__(cls)
        return cls._instance

    METADATA_KEYWORDS = True
    METADATA_ATTRIBUTIONS = True
    METADATA_FEES = True
    METADATA_ACCESS_CONSTRAINTS = True
    METADATA_CONTACT_INFO = True

    @override
    @property
    def default_abstract(self) -> str | None:
        return ""

    @property
    def active_products(self) -> Iterable[OWSNamedLayer]:
        return filter(lambda x: not x.hide, self.layer_index.values())

    @property
    def active_product_index(self) -> dict[str, OWSNamedLayer]:
        return {prod.name: prod for prod in self.active_products}

    def __init__(self, refresh=False, cfg: CFG_DICT | None = None,
                 ignore_msgfile=False, called_from_update_ranges=False) -> None:
        self.called_from_update_ranges = called_from_update_ranges
        if not self.initialised or refresh:
            self.msgfile = None
            if not cfg:
                cfg = read_config()
            super().__init__(cfg)
            try:
                self.parse_global(cast(CFG_DICT, cfg["global"]), ignore_msgfile)
            except KeyError as e:
                raise ConfigException(
                    "Missing required config entry in 'global' section: %s" % str(e)
                )

            if self.wms or self.wmts:
                self.parse_wms(cast(CFG_DICT, cfg.get("wms", {})))
            else:
                self.parse_wms({})

            if self.wcs:
                try:
                    self.parse_wcs(cast(CFG_DICT, cfg.get("wcs")))
                except KeyError as e:
                    raise ConfigException(
                        "Missing required config entry in 'wcs' section (with WCS enabled): %s" % str(e)
                    )
            else:
                self.parse_wcs(None)
            try:
                self.parse_layers(cast(list[CFG_DICT], cfg["layers"]))
            except KeyError as e:
                raise ConfigException("Missing required config entry in 'layers' section")

            try:
                if self.wmts:
                    self.parse_wmts(cast(CFG_DICT, cfg.get("wmts", {})))
                else:
                    self.parse_wmts({})
            except KeyError as e:
                raise ConfigException(
                    "Missing required config entry in 'wmts' section (with WCS enabled): %s" % str(e)
                )
            self.catalog: Catalog | None = None
            self.initialised = True
            self.declare_unready("dc")
            self.declare_unready("crses")
            self.declare_unready("native_product_index")
            self.declare_unready("all_dcs")

    #pylint: disable=attribute-defined-outside-init
    @override
    def make_ready(self, *args: Any, **kwargs: Any) -> None:
        try:
            self.dc: Datacube = Datacube(env=self.default_env, app=self.odc_app)
        except Exception as e:
            _LOG.error("ODC initialisation of env %s failed: %s", self.default_env._name, str(e))
            raise ODCInitException(e)
        if self.msg_file_name:
            try:
                with open(self.msg_file_name, "rb") as fp:
                    self.set_msg_src(read_po(fp, locale=self.default_locale, domain=self.message_domain))
            except FileNotFoundError:
                _LOG.warning("Message file %s does not exist - using metadata from config file", self.msg_file_name)
        else:
            self.set_msg_src(None)
        self.crses = {s: self.crs(s) for s in self.published_CRSs}
        self.native_product_index: dict[str, OWSNamedLayer] = {}
        self.root_layer_folder.make_ready(*args, **kwargs)
        self.all_dcs: dict[str, Datacube] = {}
        for layer in self.active_products:
            if layer.local_env._name not in self.all_dcs:
                self.all_dcs[layer.local_env._name] = layer.dc

        super().make_ready(*args, **kwargs)

    def export_metadata(self) -> Catalog:
        if self.catalog is None:
            now = datetime.datetime.now()
            header: str = f"""# Translations for datacube-ows metadata instance:
#      {self.title}
#
"""
            if self.contact_info:
                header += """# {self.contact_info.organisation} {now.isoformat()}
#"""
            else:
                header += """# {now.isoformat()}
            #"""
            self.catalog = Catalog(locale=self.default_locale,
                                   domain=self.message_domain,
                                   header_comment=header,
                                   project=self.title,
                                   version=f"{now.isoformat()}",
                                   copyright_holder=self.contact_info.organisation if self.contact_info else None,
                                   msgid_bugs_address=self.contact_info.email if self.contact_info else None,
                                   creation_date=now,
                                   revision_date=now,
                                   fuzzy=False)
            for k, v in self._metadata_registry.items():
                if self._inheritance_registry[k]:
                    continue
                if k in [
                        "folder.ows_root_hidden.title",
                        "folder.ows_root_hidden.abstract",
                        "folder.ows_root_hidden.local_keywords",
                ]:
                    continue
                self.catalog.add(id=k, string=v, auto_comments=[v])
        return self.catalog

    def parse_global(self, cfg: CFG_DICT, ignore_msgfile: bool) -> None:
        default_env = cast(str, cfg.get("env"))
        self.default_env = ODCConfig.get_environment(env=default_env)
        self.odc_app = cast(str, cfg.get("odc_app", "datacube-ows"))
        self._response_headers = cast(dict[str, str], cfg.get("response_headers", {}))
        services = cast(dict[str, bool], cfg.get("services", {}))
        self.wms = services.get("wms", True)
        self.wmts = services.get("wmts", True)
        self.wcs = services.get("wcs", False)
        if not self.wms and not self.wmts and not self.wcs:
            raise ConfigException("At least one service must be active.")
        self.locales = cast(list[str], cfg.get("supported_languages", ["en"]))
        if len(self.locales) < 1:
            raise ConfigException("You must support at least one language.")
        self.default_locale = self.locales[0]
        self.message_domain = cast(str, cfg.get("message_domain", "ows_cfg"))
        self.translations_dir = cast(str | None, cfg.get("translations_directory"))
        self.internationalised = self.translations_dir and len(self.locales) > 1
        if self.internationalised:
            _LOG.info("Internationalisation enabled.")
        if ignore_msgfile:
            self.msg_file_name: str | None = None
        else:
            self.msg_file_name = cast(str | None, cfg.get("message_file"))
        self.parse_metadata(cfg)
        self.allowed_urls = cast(str | list[str], cfg["allowed_urls"])
        self.info_url = cfg["info_url"]
        self.contact_info = ContactInfo.parse(cfg.get("contact_info"), self)
        self.attribution = AttributionCfg.parse(cast(CFG_DICT | None, cfg.get("attribution")), self)

        def make_gml_name(name: str) -> str:
            if name.startswith("EPSG:"):
                return f"http://www.opengis.net/def/crs/EPSG/0/{name[5:]}"
            else:
                return name

        self.published_CRSs: dict[str, CFG_DICT] = {}
        self.internal_CRSs: dict[str, CFG_DICT] = {}
        CRS_aliases: dict[str, CFG_DICT] = {}
        geographic_CRSs: list[str] = []
        for crs_str, crsdef in cast(dict[str, CFG_DICT], cfg["published_CRSs"]).items():
            if "alias" in crsdef:
                CRS_aliases[crs_str] = crsdef
                continue
            self.internal_CRSs[crs_str] = {
                "geographic": crsdef["geographic"],
                "horizontal_coord": crsdef.get("horizontal_coord", "longitude"),
                "vertical_coord": crsdef.get("vertical_coord", "latitude"),
                "vertical_coord_first": crsdef.get("vertical_coord_first", False),
                "gml_name": make_gml_name(crs_str),
                "alias_of": None
            }
            if crsdef["geographic"]:
                geographic_CRSs.append(crs_str)
            self.published_CRSs[crs_str] = self.internal_CRSs[crs_str]
            if self.published_CRSs[crs_str]["geographic"]:
                if self.published_CRSs[crs_str]["horizontal_coord"] != "longitude":
                    raise ConfigException(f"Published CRS {crs_str} is geographic"
                                    "but has a horizontal coordinate that is not 'longitude'")
                if self.published_CRSs[crs_str]["vertical_coord"] != "latitude":
                    raise ConfigException(f"Published CRS {crs_str} is geographic"
                                    "but has a vertical coordinate that is not 'latitude'")
        # default_geographic_CRS is used by WCS1
        if not self.wcs:
            self.default_geographic_CRS = ""
        elif not geographic_CRSs:
            raise ConfigException("At least one geographic CRS must be supplied")
        elif "EPSG:4326" in geographic_CRSs or "WGS-84" in geographic_CRSs:
            self.default_geographic_CRS = "urn:ogc:def:crs:OGC:1.3:CRS84"
        else:
            self.default_geographic_CRS = geographic_CRSs[0]

        if "EPSG:3857" not in self.published_CRSs:
            _LOG.warning("EPSG:3857 (Web mercator) is not a published CRS")
        elif "EPSG:3832" not in self.published_CRSs:
            # Have web merc but not Pacific Web merc - just add it silently.
            self.published_CRSs["EPSG:3832"] = {
                "geographic": False,
                "horizontal_coord": "x",
                "vertical_coord": "y",
                "vertical_coord_first": False,
                "gml_name": make_gml_name("EPSG:3832"),
                "alias_of": None
            }

        for alias, alias_def in CRS_aliases.items():
            target_crs = cast(str, alias_def["alias"])
            if target_crs not in self.published_CRSs:
                _LOG.warning("CRS %s defined as alias for %s, which is not a published CRS - skipping",
                             alias, target_crs)
                continue
            target_def = self.published_CRSs[target_crs]
            self.published_CRSs[alias] = target_def.copy()
            self.published_CRSs[alias]["gml_name"] = make_gml_name(alias)
            self.published_CRSs[alias]["alias_of"] = target_crs

    def parse_wms(self, cfg: CFG_DICT) -> None:
        if not self.wms and not self.wmts:
            cfg = {}
        self.s3_bucket = cast(str, cfg.get("s3_bucket", ""))
        self.s3_url = cast(str, cfg.get("s3_url", ""))
        self.s3_aws_zone = cast(str, cfg.get("s3_aws_zone", ""))
        try:
            self.wms_max_width = int(cast(str | int, cfg.get("max_width", 256)))
            self.wms_max_height = int(cast(str | int, cfg.get("max_height", 256)))
        except ValueError:
            raise ConfigException(
                f"max_width and max_height in wms section must be integers: {cfg.get('max_width', 256)},{cfg.get('max_height', 256)}"
            )
        if self.wms_max_width < 1 or self.wms_max_height < 1:
            raise ConfigException(
                f"max_width and max_height in wms section must be positive integers: {cfg.get('max_width', 256)},{cfg.get('max_height', 256)}"
            )
        self.authorities = cast(dict[str, str], cfg.get("authorities", {}))
        self.user_band_math_extension = cfg.get("user_band_math_extension", False)
        self.wms_cap_cache_age = parse_cache_age(cfg, "caps_cache_maxage", "wms")
        if "attribution" in cfg:
            _LOG.warning("Attribution entry in top level 'wms' section will be ignored. Attribution should be moved to the 'global' section")

    def parse_wcs(self, cfg: CFG_DICT | None) -> None:
        if self.wcs:
            if not isinstance(cfg, Mapping):
                raise ConfigException("WCS section missing (and WCS is enabled)")
            self.wcs_formats = WCSFormat.from_cfg(cast(dict[str, CFG_DICT], cfg["formats"]))
            self.wcs_formats_by_name = {
                fmt.name: fmt
                for fmt in self.wcs_formats
            }
            self.wcs_formats_by_mime = {
                fmt.mime: fmt
                for fmt in self.wcs_formats
            }
            if not self.wcs_formats:
                raise ConfigException("Must configure at least one wcs format to support WCS.")

            self.native_wcs_format = cfg["native_format"]
            if self.native_wcs_format not in self.wcs_formats_by_name:
                raise ConfigException(f"Configured native WCS format ({self.native_wcs_format}) not a supported format.")
            self.wcs_tiff_statistics = cfg.get("calculate_tiff_statistics", True)
            self.wcs_cap_cache_age = parse_cache_age(cfg, "caps_cache_maxage", "wcs")
            self.wcs_default_descov_age = parse_cache_age(cfg, "default_desc_cache_maxage", "wcs")
        else:
            self.wcs_formats = []
            self.wcs_formats_by_name = {}
            self.wcs_formats_by_mime = {}
            self.native_wcs_format = None
            self.wcs_tiff_statistics = False
            self.wcs_cap_cache_age = 0
            self.wcs_default_descov_age = 0

    def parse_wmts(self, cfg: CFG_DICT) -> None:
        tms_cfgs = cast(dict[str, CFG_DICT], TileMatrixSet.default_tm_sets.copy())
        if "tile_matrix_sets" in cfg:
            for identifier, tms in cast(dict[str, CFG_DICT], cfg["tile_matrix_sets"]).items():
                tms_cfgs[identifier] = tms
        self.tile_matrix_sets: dict[str, TileMatrixSet] = {}
        for identifier, tms in tms_cfgs.items():
            if len(identifier.split()) != 1:
                raise ConfigException(f"Invalid identifier: {identifier}")
            if identifier in self.tile_matrix_sets:
                raise ConfigException(f"Tile matrix set identifiers must be unique: {identifier}")
            self.tile_matrix_sets[identifier] = TileMatrixSet(identifier, tms, self)

    def parse_layers(self, cfg: list[CFG_DICT]) -> None:
        self.folder_index: dict[str, OWSFolder] = {}
        self.layer_index: dict[str, OWSNamedLayer] = {}
        self.declare_unready("native_product_index")
        self.root_layer_folder = OWSFolder(cast(CFG_DICT, {
            "title": self.title,
            "abstract": self.abstract,
            "label": "ows_root",
            "layers": cfg
        }), global_cfg=self, parent_layer=None)

    @property
    def layers(self) -> list[OWSLayer]:
        # Multiple top-level are not consistent with a strict reading of the OWS standard.
        # If we have multiple top-level folders, wrap them in an auto-generated top-level folder.
        if len(self.root_layer_folder.child_layers) == 1:
            return self.root_layer_folder.child_layers
        return [self.root_layer_folder]

    def alias_bboxes(self, bboxes: CFG_DICT) -> CFG_DICT:
        out: CFG_DICT = {}
        for crsid, crsdef in self.published_CRSs.items():
            a_crsid = cast(str, crsdef["alias_of"])
            if a_crsid:
                if a_crsid in bboxes:
                    out[crsid] = bboxes[a_crsid]
            else:
                if crsid in bboxes:
                    out[crsid] = bboxes[crsid]
        return out

    def crs(self, crsid: str) -> CRS:
        if crsid not in self.published_CRSs:
            raise ConfigException(f"CRS {crsid} is not published")
        crs_def = self.published_CRSs[crsid]
        crs_alias = crs_def["alias_of"]
        if crs_alias:
            use_crs = crs_alias
        else:
            use_crs = crsid
        return CRS(use_crs)

    def response_headers(self, d: dict[str, str]) -> dict[str, str]:
        hdrs = self._response_headers.copy()
        hdrs.update(d)
        return hdrs

def get_config(refresh=False, called_from_update_ranges=False) -> OWSConfig:
    cfg = OWSConfig(refresh=refresh, called_from_update_ranges=called_from_update_ranges)
    if not cfg.ready:
        try:
            cfg.make_ready()
        except ODCInitException:
            pass
    return cfg



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import datetime
import logging
from collections import OrderedDict
from typing import cast
from typing_extensions import override
from collections.abc import Iterable, Mapping
from uuid import UUID

import datacube
import numpy
import xarray
from odc.geo.geobox import GeoBox
from odc.geo.geom import Geometry, CRS
from odc.geo.warp import Resampling

from datacube_ows.ogc_exceptions import WMSException
from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.startup_utils import CredentialManager
from datacube_ows.styles import StyleDef
from datacube_ows.utils import log_call
from datacube_ows.wms_utils import solar_correct_data

_LOG: logging.Logger = logging.getLogger(__name__)


class ProductBandQuery:
    def __init__(self,
                 products: list[datacube.model.Product],
                 bands: Iterable[str],
                 main: bool = False, manual_merge: bool = False, ignore_time: bool = False,
                 fuse_func: datacube.api.core.FuserFunction | None = None
    ) -> None:
        self.products = products
        self.bands = set(bands)
        self.manual_merge = manual_merge
        self.fuse_func = fuse_func
        self.ignore_time = ignore_time
        self.main = main
        self.key = (
            tuple(p.id for p in self.products),
            tuple(bands)
        )

    @override
    def __str__(self) -> str:
        return f"Query bands {self.bands} from products {self.products}"

    @override
    def __hash__(self) -> int:
        return hash(self.key)

    @classmethod
    def style_queries(cls, style: StyleDef, resource_limited: bool = False) -> list["ProductBandQuery"]:
        queries = [
            cls.simple_layer_query(style.product, style.needed_bands,
                                   manual_merge=style.product.data_manual_merge,
                                   fuse_func=style.product.fuse_func,
                                   resource_limited=resource_limited)
        ]
        for fp in style.flag_products:
            if fp.products_match(style.product.product_names):
                for band in fp.bands:
                    assert band in style.needed_bands, "Style band not in needed bands list"
            else:
                if resource_limited:
                    pq_products = fp.low_res_products
                else:
                    pq_products = fp.products
                queries.append(cls(
                    pq_products,
                    list(fp.bands),
                    manual_merge=fp.manual_merge,
                    ignore_time=fp.ignore_time,
                    fuse_func=fp.fuse_func
                ))
        return queries

    @classmethod
    def full_layer_queries(cls,
                           layer: OWSNamedLayer,
                           main_bands: list[str] | None = None) -> list["ProductBandQuery"]:
        if main_bands:
            needed_bands: Iterable[str] = main_bands
        else:
            needed_bands = set(layer.band_idx.band_cfg.keys())
        queries = [
            cls.simple_layer_query(layer, needed_bands,
                                   manual_merge=layer.data_manual_merge,
                                   fuse_func=layer.fuse_func,
                                   resource_limited=False)
        ]
        for fpb in layer.allflag_productbands:
            if fpb.products_match(layer.product_names):
                for band in fpb.bands:
                    assert band in needed_bands, "main product band not in needed bands list"
            else:
                pq_products = fpb.products
                queries.append(cls(
                    pq_products,
                    list(fpb.bands),
                    manual_merge=fpb.manual_merge,
                    ignore_time=fpb.ignore_time,
                    fuse_func=fpb.fuse_func
                ))
        return queries

    @classmethod
    def simple_layer_query(cls, layer: OWSNamedLayer,
                           bands: Iterable[str],
                           manual_merge: bool = False,
                           fuse_func: datacube.api.core.FuserFunction | None = None,
                           resource_limited: bool = False) -> "ProductBandQuery":
        if resource_limited:
            main_products = layer.low_res_products
        else:
            main_products = layer.products
        return cls(main_products, bands, manual_merge=manual_merge, main=True, fuse_func=fuse_func)

PerPBQReturnType = xarray.DataArray | Iterable[UUID]

class DataStacker:
    @log_call
    def __init__(self,
                 layer: OWSNamedLayer,
                 geobox: GeoBox,
                 times: list[datetime.datetime],
                 resampling: Resampling | None = None,
                 style: StyleDef | None = None,
                 bands: list[str] | None = None) -> None:
        self._layer = layer
        self.cfg = layer.global_cfg
        self._geobox = geobox
        self._resampling = resampling if resampling is not None else "nearest"
        self.style = style
        if style:
            self._needed_bands = list(style.needed_bands)
        elif bands:
            self._needed_bands = [self._layer.band_idx.locale_band(b) for b in bands]
        else:
            self._needed_bands = list(self._layer.band_idx.measurements.keys())

        for band in self._layer.always_fetch_bands:
            if band not in self._needed_bands:
                self._needed_bands.append(band)
        self.raw_times = times
        if self._layer.mosaic_date_func:
            self._times = [self._layer.mosaic_date_func(layer.ranges.times)]
        else:
            self._times = [
                    self._layer.search_times(
                            t, self._geobox)
                    for t in times
            ]
        self.group_by = self._layer.dataset_groupby()
        self.resource_limited = False

    def needed_bands(self) -> list[str]:
        return self._needed_bands

    def n_datasets(self) -> int:
        if self.style:
            # we have a style - lets go with that.
            queries = ProductBandQuery.style_queries(self.style)
        else:
            # Just take needed bands.
            queries = [ProductBandQuery.simple_layer_query(self._layer, self.needed_bands())]
        geom = self._geobox.extent
        for query in queries:
            if query.ignore_time:
                qry_times = None
            else:
                qry_times = self._times
            return self._layer.ows_index().count(self._layer, times=qry_times, geom=geom, products=query.products)
        return 0

    def extent(self, crs: CRS | None = None) -> Geometry | None:
        query = ProductBandQuery.simple_layer_query(
                self._layer,
                self.needed_bands(),
                self.resource_limited
        )
        geom = self._geobox.extent
        if query.ignore_time:
            times = None
        else:
            times = self._times
        return self._layer.ows_index().extent(self._layer, times=times, geom=geom, products=query.products, crs=crs)

    def dsids(self) -> dict[ProductBandQuery, Iterable[UUID]]:
        if self.style:
            # we have a style - lets go with that.
            queries = ProductBandQuery.style_queries(self.style)
        else:
            # Just take needed bands.
            queries = [ProductBandQuery.simple_layer_query(self._layer, self.needed_bands())]
        results: list[tuple[ProductBandQuery, Iterable[UUID]]] = []
        for query in queries:
            if query.ignore_time:
                qry_times = None
            else:
                qry_times = self._times
            result = self._layer.ows_index().dsid_search(self._layer, times=qry_times, geom=self._geobox.extent,
                                                         products=query.products)
            results.append((query, result))
        return OrderedDict(results)

    def datasets_all_time(self, point: Geometry | None = None) -> xarray.DataArray:
        query = ProductBandQuery.simple_layer_query(
                    self._layer,
                    self.needed_bands(),
                    self.resource_limited)
        if point:
            geom = point
        else:
            geom = self._geobox.extent
        result = self._layer.ows_index().ds_search(
            layer=self._layer,
            geom=geom,
            products=query.products)
        grpd_result = datacube.Datacube.group_datasets(result, self.group_by)
        return grpd_result

    def datasets(self,
                 all_flag_bands: bool = False,
                 point: Geometry | None = None,
                 ) -> dict[ProductBandQuery, xarray.DataArray]:
        if self.style:
            # we have a style - lets go with that.
            queries = ProductBandQuery.style_queries(self.style)
        elif all_flag_bands:
            queries = ProductBandQuery.full_layer_queries(self._layer, self.needed_bands())
        else:
            # Just take needed bands.
            queries = [ProductBandQuery.simple_layer_query(self._layer, self.needed_bands())]

        if point:
            geom = point
        else:
            geom = self._geobox.extent
        results: list[tuple[ProductBandQuery, xarray.DataArray]] = []
        for query in queries:
            if query.ignore_time:
                qry_times = None
            else:
                qry_times = self._times
            result = self._layer.ows_index().ds_search(
                               layer=self._layer,
                               times=qry_times,
                               geom=geom,
                               products=query.products)
            grpd_result = datacube.Datacube.group_datasets(result, self.group_by)
            results.append((query, grpd_result))
        return OrderedDict(results)

    def create_nodata_filled_flag_bands(self, data: xarray.Dataset, pbq: ProductBandQuery) -> xarray.Dataset:
        var = None
        for var in data.data_vars.variables.keys():
            break
        if var is None:
            raise WMSException("Cannot add default flag data as there is no non-flag data available")
        template = cast(xarray.DataArray, getattr(data, cast(str, var)))
        data_new_bands = {}
        for band in pbq.bands:
            default_value = pbq.products[0].measurements[band].nodata
            new_data: numpy.ndarray = numpy.ndarray(template.shape, dtype="uint8")
            new_data.fill(default_value)
            qry_result = template.copy(data=new_data)
            data_new_bands[band] = qry_result
        data = data.assign(data_new_bands)
        for band in pbq.bands:
            data[band].attrs["flags_definition"] = pbq.products[0].measurements[band].flags_definition
        return data

    @log_call
    def data(self,
             datasets_by_query: dict[ProductBandQuery, xarray.DataArray],
             skip_corrections=False) -> xarray.Dataset | None:
        # pylint: disable=too-many-locals, consider-using-enumerate
        # datasets is an XArray DataArray of datasets grouped by time.
        data: xarray.Dataset | None = None
        for pbq, datasets in datasets_by_query.items():
            if data is not None and len(data.time) == 0:
                # No data, so no need for masking data.
                continue
            measurements = pbq.products[0].lookup_measurements(pbq.bands)
            fuse_func = pbq.fuse_func
            if pbq.manual_merge:
                qry_result = self.manual_data_stack(datasets, measurements, pbq.bands, skip_corrections, fuse_func=fuse_func)
            else:
                qry_result = self.read_data(datasets, measurements, self._geobox, resampling=self._resampling, fuse_func=fuse_func)
            if qry_result is None:
                continue
            if data is None:
                data = qry_result
                continue
            if len(data.time) == 0:
                # No data, so no need for masking data.
                continue
            if pbq.ignore_time:
                # regularise time dimension:
                if len(qry_result.time) > 1:
                    raise WMSException("Cannot ignore time on PQ (flag) bands from a time-aware product")
                elif len(qry_result.time) == len(data.time):
                    qry_result["time"] = data.time
                else:
                    if len(qry_result.time) == 0:
                        data = self.create_nodata_filled_flag_bands(data, pbq)
                        continue
                    else:
                        data_new_bands = {}
                        for band in pbq.bands:
                            band_data = qry_result[band]
                            timeless_band_data = band_data.sel(time=qry_result.time.values[0])
                            band_time_slices = []
                            for dt in data.time.values:
                                band_time_slices.append(timeless_band_data)
                            timed_band_data = xarray.concat(band_time_slices, data.time)
                            data_new_bands[band] = timed_band_data
                    data = data.assign(data_new_bands)
                    continue
            elif len(qry_result.time) == 0:
                # Time-aware mask product has no data, but main product does.
                data = self.create_nodata_filled_flag_bands(data, pbq)
                continue
            assert data is not None
            qry_result.coords["time"] = data.coords["time"]
            data = cast(xarray.Dataset, xarray.combine_by_coords([data, qry_result], join="exact"))

        return data

    @log_call
    def manual_data_stack(self,
                          datasets: xarray.DataArray,
                          measurements: Mapping[str, datacube.model.Measurement],
                          bands: set[str],
                          skip_corrections: bool,
                          fuse_func: datacube.api.core.FuserFunction | None) -> xarray.Dataset | None:
        # pylint: disable=too-many-locals, too-many-branches
        # manual merge
        if self.style:
            flag_bands: Iterable[str] = set(filter(lambda b: b in self.style.flag_bands, bands))  # type: ignore[arg-type]
            non_flag_bands: Iterable[str] = set(filter(lambda b: b not in self.style.flag_bands, bands))  #type: ignore[arg-type]
        else:
            non_flag_bands = bands
            flag_bands = set()
        time_slices = []
        for dt in datasets.time.values:
            tds = datasets.sel(time=dt)
            merged = None
            for ds in cast(Iterable[datacube.model.Dataset], tds.values.item()):
                d = self.read_data_for_single_dataset(ds, measurements, self._geobox, fuse_func=fuse_func)
                extent_mask = None
                for band in non_flag_bands:
                    for f in self._layer.extent_mask_func:
                        if extent_mask is None:
                            extent_mask = f(d, band)
                        else:
                            extent_mask &= f(d, band)
                if extent_mask is not None:
                    d = d.where(extent_mask)
                if self._layer.solar_correction and not skip_corrections:
                    for band in non_flag_bands:
                        d[band] = solar_correct_data(d[band], ds)
                if merged is None:
                    merged = d
                else:
                    merged = merged.combine_first(d)
            if merged is None:
                continue
            for band in flag_bands:
                # REVISIT: not sure about type converting one band like this?
                merged[band] = merged[band].astype('uint16', copy=True)
                merged[band].attrs = d[band].attrs
            time_slices.append(merged)

        if not time_slices:
            return None
        result = xarray.concat(time_slices, datasets.time)
        return result

    # Read data for given datasets and measurements per the output_geobox
    # TODO: Make skip_broken passed in via config
    @log_call
    def read_data(self,
                  datasets: xarray.DataArray,
                  measurements: Mapping[str, datacube.model.Measurement],
                  geobox: GeoBox,
                  skip_broken: bool = True,
                  resampling: Resampling = "nearest",
                  fuse_func: datacube.api.core.FuserFunction | None = None) -> xarray.Dataset:
        CredentialManager.check_cred()
        try:
            return datacube.Datacube.load_data(
                    datasets,
                    geobox,
                    measurements=measurements,
                    fuse_func=fuse_func,
                    skip_broken_datasets=skip_broken,
                    patch_url=self._layer.patch_url,
                    resampling=resampling)
        except Exception as e:
            _LOG.error("Error (%s) in load_data: %s", e.__class__.__name__, str(e))
            raise

    # TODO: Make skip_broken passed in via config
    @log_call
    def read_data_for_single_dataset(self,
                                     dataset: datacube.model.Dataset,
                                     measurements: Mapping[str, datacube.model.Measurement],
                                     geobox: GeoBox,
                                     skip_broken: bool = True,
                                     resampling: Resampling = "nearest",
                                     fuse_func: datacube.api.core.FuserFunction | None = None) -> xarray.Dataset:
        datasets = [dataset]
        dc_datasets = datacube.Datacube.group_datasets(datasets, self._layer.time_resolution.dataset_groupby())
        CredentialManager.check_cred()
        try:
            return datacube.Datacube.load_data(
                dc_datasets,
                geobox,
                measurements=measurements,
                fuse_func=fuse_func,
                skip_broken_datasets=skip_broken,
                patch_url=self._layer.patch_url,
                resampling=resampling)
        except Exception as e:
            _LOG.error("Error (%s) in load_data: %s", e.__class__.__name__, str(e))
            raise
